import pytest

from nanomanifold import SE3, SO3


def test_so3_weighted_mean_propagates_torch_gradients():
    torch = pytest.importorskip("torch")

    q1 = torch.tensor([0.9238795325, 0.3826834324, 0.0, 0.0], dtype=torch.float64, requires_grad=True)
    q2 = torch.tensor([0.9238795325, 0.0, 0.3826834324, 0.0], dtype=torch.float64, requires_grad=True)
    weights = torch.tensor([0.2, 0.8], dtype=torch.float64)

    out = SO3.weighted_mean([q1, q2], weights, xp=torch)
    loss = (out * torch.tensor([0.7, -0.2, 0.5, 1.1], dtype=torch.float64)).sum()
    loss.backward()

    assert q1.grad is not None
    assert q2.grad is not None
    assert torch.isfinite(q1.grad).all()
    assert torch.isfinite(q2.grad).all()
    assert torch.max(torch.abs(q1.grad)) > 0
    assert torch.max(torch.abs(q2.grad)) > 0


def test_se3_weighted_mean_propagates_rotation_torch_gradients():
    torch = pytest.importorskip("torch")

    se3_1 = torch.tensor([0.9238795325, 0.3826834324, 0.0, 0.0, 0.0, 0.0, 0.0], dtype=torch.float64, requires_grad=True)
    se3_2 = torch.tensor([0.9238795325, 0.0, 0.3826834324, 0.0, 1.0, 2.0, 3.0], dtype=torch.float64, requires_grad=True)
    weights = torch.tensor([0.2, 0.8], dtype=torch.float64)

    out = SE3.weighted_mean([se3_1, se3_2], weights, xp=torch)
    loss = (out[..., :4] * torch.tensor([0.7, -0.2, 0.5, 1.1], dtype=torch.float64)).sum()
    loss.backward()

    assert torch.isfinite(se3_1.grad[..., :4]).all()
    assert torch.isfinite(se3_2.grad[..., :4]).all()
    assert torch.max(torch.abs(se3_1.grad[..., :4])) > 0
    assert torch.max(torch.abs(se3_2.grad[..., :4])) > 0



from types import ModuleType
from typing import Any

from jaxtyping import Float

from nanomanifold.common import get_namespace
from nanomanifold.SO3 import inverse as so3_inverse
from nanomanifold.SO3 import rotate_points

from .canonicalize import canonicalize


def inverse(se3: Float[Any, "... 7"], *, xp: ModuleType | None = None) -> Float[Any, "... 7"]:
    """Compute the inverse of SE(3) transformations.

    For an SE(3) transformation T = [R, t] represented as [q, t],
    the inverse is T^(-1) = [R^T, -R^T * t] represented as [q^(-1), -q^(-1) * t].

    Args:
        se3: SE(3) transformation in [w, x, y, z, tx, ty, tz] format
        xp: Array namespace (e.g. torch, jax.numpy). If None, auto-detected.

    Returns:
        Inverse SE(3) transformation
    """
    if xp is None:
        xp = get_namespace(se3)

    se3 = canonicalize(se3, xp=xp)

    q = se3[..., :4]
    t = se3[..., 4:7]

    q_inv = so3_inverse(q, xp=xp)

    t_inv = -rotate_points(q_inv, t[..., None, :], xp=xp).squeeze(-2)

    result = xp.concatenate([q_inv, t_inv], axis=-1)

    return canonicalize(result, xp=xp)



"""Tests for SO3.random uniform sampling."""

import numpy as np
import pytest
from conftest import ATOL, TEST_BACKENDS, TEST_BATCH_DIMS, TEST_PASS_XP, TEST_PRECISIONS, get_xp_kwargs

from nanomanifold import SO3
from nanomanifold.common import get_namespace_by_name


def _get_dtype(backend, precision, pass_xp=True):
    # When xp is not passed, random() defaults to numpy, so use numpy dtypes
    if not pass_xp or backend == "numpy":
        return getattr(np, f"float{precision}")
    elif backend == "torch":
        torch = pytest.importorskip("torch")
        return getattr(torch, f"float{precision}")
    else:
        jax = pytest.importorskip("jax")
        return getattr(jax.numpy, f"float{precision}")


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_random_shape(backend, batch_dims, precision, pass_xp):
    """Output shape is (*shape, 4)."""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    dtype = _get_dtype(backend, precision, pass_xp)

    if backend == "jax" and pass_xp:
        jax = pytest.importorskip("jax")
        xp_kwargs["key"] = jax.random.PRNGKey(0)
    elif backend == "jax" and not pass_xp:
        pytest.skip("random() without xp defaults to numpy, not jax")

    q = SO3.random(*batch_dims, dtype=dtype, **xp_kwargs)
    assert q.shape == batch_dims + (4,)


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_random_unit_norm(backend, batch_dims, precision, pass_xp):
    """All quaternions have unit norm."""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    dtype = _get_dtype(backend, precision, pass_xp)

    if backend == "jax" and pass_xp:
        jax = pytest.importorskip("jax")
        xp_kwargs["key"] = jax.random.PRNGKey(1)
    elif backend == "jax" and not pass_xp:
        pytest.skip("random() without xp defaults to numpy, not jax")

    q = SO3.random(*batch_dims, dtype=dtype, **xp_kwargs)
    q_np = np.array(q)
    norms = np.linalg.norm(q_np, axis=-1)
    assert np.allclose(norms, 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_random_canonical(backend, batch_dims, precision, pass_xp):
    """All quaternions have w >= 0."""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    dtype = _get_dtype(backend, precision, pass_xp)

    if backend == "jax" and pass_xp:
        jax = pytest.importorskip("jax")
        xp_kwargs["key"] = jax.random.PRNGKey(2)
    elif backend == "jax" and not pass_xp:
        pytest.skip("random() without xp defaults to numpy, not jax")

    q = SO3.random(*batch_dims, dtype=dtype, **xp_kwargs)
    q_np = np.array(q)
    assert np.all(q_np[..., 0] >= -1e-7)


@pytest.mark.parametrize("backend", TEST_BACKENDS)
def test_random_statistical_uniformity(backend):
    """For large N, mean dot product with any fixed quaternion ~ 0."""
    N = 10000
    dtype = _get_dtype(backend, 32)
    xp_kwargs = {}
    if backend != "numpy":
        xp = get_namespace_by_name(backend)
        xp_kwargs["xp"] = xp
    if backend == "jax":
        jax = pytest.importorskip("jax")
        xp_kwargs["key"] = jax.random.PRNGKey(42)

    q = SO3.random(N, dtype=dtype, **xp_kwargs)
    q_np = np.array(q)

    # Use xyz components (not w) to test uniformity; w is biased by canonicalization
    ref = np.array([0.0, 1.0, 0.0, 0.0], dtype=np.float32)
    dots = np.sum(q_np * ref, axis=-1)
    mean_dot = np.mean(dots)
    # For uniform SO(3) sampling, mean of x component ~ 0
    assert abs(mean_dot) < 0.05, f"Mean dot product {mean_dot} too far from 0"


def test_random_jax_deterministic():
    """Same JAX key gives identical output."""
    jax = pytest.importorskip("jax")
    xp = get_namespace_by_name("jax")
    key = jax.random.PRNGKey(123)

    q1 = SO3.random(5, xp=xp, key=key)
    q2 = SO3.random(5, xp=xp, key=key)
    np.testing.assert_array_equal(np.array(q1), np.array(q2))


def test_random_jax_requires_key():
    """JAX backend raises ValueError without key."""
    pytest.importorskip("jax")
    xp = get_namespace_by_name("jax")
    with pytest.raises(ValueError, match="PRNG key"):
        SO3.random(5, xp=xp)


def test_random_jax_jit():
    """SO3.random works under jax.jit."""
    jax = pytest.importorskip("jax")
    xp = get_namespace_by_name("jax")

    @jax.jit
    def sample(key):
        return SO3.random(5, xp=xp, key=key)

    key = jax.random.PRNGKey(0)
    q = sample(key)
    assert q.shape == (5, 4)
    norms = np.linalg.norm(np.array(q), axis=-1)
    assert np.allclose(norms, 1.0, atol=1e-6)



from .canonicalize import canonicalize
from .conversions.matrix import from_matrix, to_matrix
from .conversions.rt import from_rt, to_rt
from .exp import exp
from .hat import hat
from .inverse import inverse
from .log import log
from .multiply import multiply
from .random import random
from .slerp import slerp
from .transform_points import transform_points
from .vee import vee
from .weighted_mean import mean, weighted_mean

__all__ = [
    "from_matrix",
    "to_matrix",
    "from_rt",
    "to_rt",
    "canonicalize",
    "multiply",
    "inverse",
    "transform_points",
    "log",
    "exp",
    "hat",
    "vee",
    "slerp",
    "weighted_mean",
    "mean",
    "random",
]



import numpy as np
import pytest
from conftest import (
    ATOL,
    TEST_BACKENDS,
    TEST_BATCH_DIMS,
    TEST_PASS_XP,
    TEST_PRECISIONS,
    get_xp_kwargs,
    identity_quaternion,
    random_quaternion,
)
from scipy.spatial.transform import Rotation as R

from nanomanifold import SO3


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_multiply_identity(backend, batch_dims, precision, pass_xp):
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    # Create random quaternion and identity
    quat = random_quaternion(batch_dims=batch_dims, backend=backend, precision=precision)
    identity = identity_quaternion(batch_dims=batch_dims, backend=backend, precision=precision)

    # Test identity * quat = quat
    result1 = SO3.multiply(identity, quat, **xp_kwargs)
    # Test quat * identity = quat
    result2 = SO3.multiply(quat, identity, **xp_kwargs)

    assert result1.dtype == quat.dtype
    assert result1.shape == quat.shape
    assert result2.dtype == quat.dtype
    assert result2.shape == quat.shape

    # Convert to numpy arrays and compare
    quat_np = np.array(quat)
    result1_np = np.array(result1)
    result2_np = np.array(result2)

    # Check quaternion equivalence (q and -q represent the same rotation)
    dot_products1 = np.sum(quat_np * result1_np, axis=-1)
    dot_products2 = np.sum(quat_np * result2_np, axis=-1)
    assert np.allclose(np.abs(dot_products1), 1.0, atol=ATOL[precision])
    assert np.allclose(np.abs(dot_products2), 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_multiply_inverse(backend, batch_dims, precision, pass_xp):
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    # Create random quaternion
    quat = random_quaternion(batch_dims=batch_dims, backend=backend, precision=precision)
    quat_inv = SO3.inverse(quat, **xp_kwargs)

    # Test quat * quat_inv = identity
    result1 = SO3.multiply(quat, quat_inv, **xp_kwargs)
    # Test quat_inv * quat = identity
    result2 = SO3.multiply(quat_inv, quat, **xp_kwargs)

    assert result1.dtype == quat.dtype
    assert result1.shape == quat.shape
    assert result2.dtype == quat.dtype
    assert result2.shape == quat.shape

    # Convert to numpy arrays and compare with identity quaternion
    identity = identity_quaternion(batch_dims=batch_dims, backend=backend, precision=precision)
    identity_np = np.array(identity)
    result1_np = np.array(result1)
    result2_np = np.array(result2)

    # Check quaternion equivalence (q and -q represent the same rotation)
    dot_products1 = np.sum(identity_np * result1_np, axis=-1)
    dot_products2 = np.sum(identity_np * result2_np, axis=-1)
    assert np.allclose(np.abs(dot_products1), 1.0, atol=ATOL[precision])
    assert np.allclose(np.abs(dot_products2), 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_multiply_matrix_equivalence(backend, batch_dims, precision, pass_xp):
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    # Create two random quaternions
    q1 = random_quaternion(batch_dims=batch_dims, backend=backend, precision=precision)
    q2 = random_quaternion(batch_dims=batch_dims, backend=backend, precision=precision)

    # Multiply quaternions
    q_result = SO3.multiply(q1, q2, **xp_kwargs)

    # Convert to matrices and multiply
    R1 = SO3.to_matrix(q1, **xp_kwargs)
    R2 = SO3.to_matrix(q2, **xp_kwargs)
    R_result = np.matmul(np.array(R1), np.array(R2))

    # Convert result matrix back to quaternion
    q_from_matrix = SO3.from_matrix(R_result)

    assert q_result.dtype == q1.dtype
    assert q_result.shape == q1.shape

    # Convert to numpy arrays and compare
    q_result_np = np.array(q_result)
    q_from_matrix_np = np.array(q_from_matrix)

    # Check quaternion equivalence (q and -q represent the same rotation)
    dot_products = np.sum(q_result_np * q_from_matrix_np, axis=-1)
    assert np.allclose(np.abs(dot_products), 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_multiply_scipy(backend, batch_dims, pass_xp):
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    # Create two random quaternions
    q1 = random_quaternion(batch_dims=batch_dims, backend=backend, precision=32)
    q2 = random_quaternion(batch_dims=batch_dims, backend=backend, precision=32)

    # Multiply using nanomanifold
    q_result = SO3.multiply(q1, q2, **xp_kwargs)

    # Convert to numpy for scipy
    q1_np = np.array(q1)
    q2_np = np.array(q2)

    # Convert from [w, x, y, z] to scipy's [x, y, z, w] format
    q1_scipy = np.concatenate([q1_np[..., 1:4], q1_np[..., 0:1]], axis=-1)
    q2_scipy = np.concatenate([q2_np[..., 1:4], q2_np[..., 0:1]], axis=-1)

    # Reshape for scipy processing
    q1_flat = q1_scipy.reshape(-1, 4)
    q2_flat = q2_scipy.reshape(-1, 4)

    # Multiply using scipy
    q_result_scipy = []
    for i in range(len(q1_flat)):
        r1 = R.from_quat(q1_flat[i : i + 1])
        r2 = R.from_quat(q2_flat[i : i + 1])
        # For scipy, r1 * r2 means apply r2 first, then r1 (same as our convention)
        r_result = r1 * r2
        q_result_scipy.append(r_result.as_quat())

    q_result_scipy = np.array(q_result_scipy).reshape(q1_scipy.shape)
    # Convert back to [w, x, y, z] format
    q_result_scipy = np.concatenate([q_result_scipy[..., 3:4], q_result_scipy[..., 0:3]], axis=-1)

    assert q_result.dtype == q1.dtype
    assert q_result.shape == q_result_scipy.shape

    q_result_np = np.array(q_result)
    # Check quaternion equivalence (q and -q represent the same rotation)
    dot_products = np.sum(q_result_np * q_result_scipy, axis=-1)
    assert np.allclose(np.abs(dot_products), 1.0, atol=1e-6)


@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_multiply_differentiability_torch(batch_dims):
    torch = pytest.importorskip("torch")
    # Use double precision for gradient checking as recommended by PyTorch
    dtype = torch.float64

    # Random quaternion inputs
    q1 = random_quaternion(batch_dims=batch_dims, backend="torch").to(dtype).requires_grad_(True)
    q2 = random_quaternion(batch_dims=batch_dims, backend="torch").to(dtype).requires_grad_(True)

    # Check gradients of SO3.multiply with respect to first argument
    def f_q1(q):
        return SO3.multiply(q, q2.detach())

    assert torch.autograd.gradcheck(f_q1, (q1,), eps=1e-6, atol=1e-5)

    # Check gradients of SO3.multiply with respect to second argument
    def f_q2(q):
        return SO3.multiply(q1.detach(), q)

    assert torch.autograd.gradcheck(f_q2, (q2,), eps=1e-6, atol=1e-5)


@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_multiply_jittability_jax(batch_dims):
    """Test that multiply function is JIT-compatible with JAX"""
    jax = pytest.importorskip("jax")

    # Create test quaternions
    q1 = random_quaternion(batch_dims=batch_dims, backend="jax", precision=32)
    q2 = random_quaternion(batch_dims=batch_dims, backend="jax", precision=32)

    # Define JIT-compiled function
    @jax.jit
    def jit_multiply(q1, q2):
        return SO3.multiply(q1, q2)

    # Test that JIT compilation works and compare with non-JIT
    result_jit = jit_multiply(q1, q2)
    result_non_jit = SO3.multiply(q1, q2)

    # Verify results match between JIT and non-JIT
    assert jax.numpy.allclose(result_jit, result_non_jit, atol=1e-6)

    # Verify result has correct shape and is unit quaternion
    assert result_jit.shape == batch_dims + (4,)
    result_norm = jax.numpy.linalg.norm(result_jit, axis=-1)
    assert jax.numpy.allclose(result_norm, 1.0, atol=1e-6)


def test_multiply_associativity():
    """Test that quaternion multiplication is associative: (q1 * q2) * q3 = q1 * (q2 * q3)"""
    # Create three random quaternions
    q1 = random_quaternion(batch_dims=(), backend="numpy", precision=32)
    q2 = random_quaternion(batch_dims=(), backend="numpy", precision=32)
    q3 = random_quaternion(batch_dims=(), backend="numpy", precision=32)

    # Compute (q1 * q2) * q3
    temp1 = SO3.multiply(q1, q2)
    result1 = SO3.multiply(temp1, q3)

    # Compute q1 * (q2 * q3)
    temp2 = SO3.multiply(q2, q3)
    result2 = SO3.multiply(q1, temp2)

    # Convert to numpy arrays and compare
    result1_np = np.array(result1)
    result2_np = np.array(result2)

    # Check quaternion equivalence (q and -q represent the same rotation)
    dot_product = np.sum(result1_np * result2_np, axis=-1)
    assert np.allclose(np.abs(dot_product), 1.0, atol=1e-6)


def test_multiply_specific_rotations():
    """Test multiplication of specific known rotations."""
    # 90-degree rotation around z-axis: [cos(π/4), 0, 0, sin(π/4)]
    sqrt2_inv = 1.0 / np.sqrt(2.0)
    q_z90 = np.array([sqrt2_inv, 0, 0, sqrt2_inv])

    # 90-degree rotation around x-axis: [cos(π/4), sin(π/4), 0, 0]
    q_x90 = np.array([sqrt2_inv, sqrt2_inv, 0, 0])

    # Multiply: first rotate around z, then around x
    q_result = SO3.multiply(q_x90, q_z90)

    # Verify by checking what happens to the point (1, 0, 0)
    point = np.array([[1.0, 0.0, 0.0]])

    # Expected: (1,0,0) -> z90 -> (0,1,0) -> x90 -> (0,0,1)
    rotated = SO3.rotate_points(q_result, point)
    rotated_np = np.array(rotated)
    expected = np.array([[0.0, 0.0, 1.0]])

    assert np.allclose(rotated_np, expected, atol=1e-6)


def test_multiply_broadcasting():
    """Test broadcasting between different batch sizes."""
    # Single quaternion and batch of quaternions
    q_single = random_quaternion(batch_dims=(), backend="numpy", precision=32)
    q_batch = random_quaternion(batch_dims=(3,), backend="numpy", precision=32)

    # Test broadcasting
    result1 = SO3.multiply(q_single, q_batch)
    result2 = SO3.multiply(q_batch, q_single)

    assert result1.shape == (3, 4)
    assert result2.shape == (3, 4)

    # Results should be different due to non-commutative multiplication
    result1_np = np.array(result1)
    result2_np = np.array(result2)

    # They shouldn't be equal (quaternion multiplication is not commutative)
    # Just check they have the right shapes and are valid quaternions
    norms1 = np.linalg.norm(result1_np, axis=-1)
    norms2 = np.linalg.norm(result2_np, axis=-1)
    assert np.allclose(norms1, 1.0, atol=1e-6)
    assert np.allclose(norms2, 1.0, atol=1e-6)



import math
from types import ModuleType
from typing import Any

from jaxtyping import Float

from nanomanifold import common
from nanomanifold.common import get_namespace

from .quaternion import canonicalize


def to_axis_angle(q: Float[Any, "... 4"], *, xp: ModuleType | None = None) -> Float[Any, "... 3"]:
    if xp is None:
        xp = get_namespace(q)
    q_canonical = canonicalize(q, xp=xp)

    w = q_canonical[..., 0:1]
    xyz = q_canonical[..., 1:4]

    norm_xyz = xp.linalg.norm(xyz, axis=-1, keepdims=True)

    thresh = xp.asarray(math.sqrt(common.safe_eps(q.dtype, xp, scale=1.0)), dtype=q.dtype)
    small_angle_mask = norm_xyz < thresh

    axis_angle_small = 2.0 * xyz

    w_clipped = xp.clip(w, 0.0, 1.0)
    angle = 2 * xp.atan2(norm_xyz, w_clipped)

    safe_norm = xp.where(norm_xyz < thresh, xp.ones_like(norm_xyz), norm_xyz)
    axis = xyz / safe_norm

    axis_angle_large = angle * axis

    axis_angle = xp.where(small_angle_mask, axis_angle_small, axis_angle_large)

    return axis_angle


def from_axis_angle(axis_angle: Float[Any, "... 3"], *, xp: ModuleType | None = None) -> Float[Any, "... 4"]:
    if xp is None:
        xp = get_namespace(axis_angle)

    angle = xp.linalg.norm(axis_angle, axis=-1)

    thresh = xp.asarray(math.sqrt(common.safe_eps(axis_angle.dtype, xp, scale=1.0)), dtype=axis_angle.dtype)
    small_angle_mask = angle < thresh
    safe_angle = xp.where(small_angle_mask, xp.ones_like(angle), angle)
    axis = axis_angle / safe_angle[..., None]

    half_angle = angle / 2

    # For small angles, use Taylor series: cos(x) ≈ 1 - x²/2, sin(x) ≈ x
    cos_half = xp.cos(half_angle)
    sin_half = xp.sin(half_angle)

    w = cos_half[..., None]

    xyz_normal = sin_half[..., None] * axis
    xyz_small = axis_angle / 2.0

    xyz = xp.where(small_angle_mask[..., None], xyz_small, xyz_normal)

    q = xp.concatenate([w, xyz], axis=-1)

    mask = q[..., 0:1] < 0
    q = xp.where(mask, -q, q)

    return q



"""Tests for SO3.conversions pairwise wrappers."""

import numpy as np
import pytest
from conftest import ATOL, TEST_BACKENDS, TEST_BATCH_DIMS, TEST_PASS_XP, get_xp_kwargs, random_quaternion

from nanomanifold import SO3

# ---------------------------------------------------------------------------
# Helpers to build inputs for each source representation
# ---------------------------------------------------------------------------

_EULER_CONVENTIONS = ["ZYX", "XYZ", "ZXZ"]


def _make_input(source: str, batch_dims, backend, precision=32, convention="ZYX"):
    """Return a sample array in the given representation."""
    q = random_quaternion(batch_dims=batch_dims, backend=backend, precision=precision)
    if source == "axis_angle":
        return SO3.to_axis_angle(q)
    if source == "euler":
        return SO3.to_euler(q, convention=convention)
    if source == "matrix":
        return SO3.to_matrix(q)
    if source == "quat_wxyz":
        return q
    if source == "quat_xyzw":
        return SO3.to_quat_xyzw(q)
    if source == "sixd":
        return SO3.to_6d(q)
    raise ValueError(source)


def _manual_convert(x, source, target, convention="ZYX"):
    """Convert via the two-step from/to primitive path (the reference)."""
    # source -> quat_wxyz
    if source == "axis_angle":
        q = SO3.from_axis_angle(x)
    elif source == "euler":
        q = SO3.from_euler(x, convention=convention)
    elif source == "matrix":
        q = SO3.from_matrix(x)
    elif source == "quat_wxyz":
        q = SO3.canonicalize(x)
    elif source == "quat_xyzw":
        q = SO3.from_quat_xyzw(x)
    elif source == "sixd":
        q = SO3.from_6d(x)
    else:
        raise ValueError(source)

    # quat_wxyz -> target
    if target == "axis_angle":
        return SO3.to_axis_angle(q)
    if target == "euler":
        return SO3.to_euler(q, convention=convention)
    if target == "matrix":
        return SO3.to_matrix(q)
    if target == "quat_wxyz":
        return SO3.canonicalize(q)
    if target == "quat_xyzw":
        return SO3.to_quat_xyzw(q)
    if target == "sixd":
        return SO3.to_6d(q)
    raise ValueError(target)


# ---------------------------------------------------------------------------
# All 30 source-target pairs
# ---------------------------------------------------------------------------

_REPS = ["axis_angle", "euler", "matrix", "quat_wxyz", "quat_xyzw", "sixd"]
_PAIRS = [(s, t) for s in _REPS for t in _REPS if s != t]


def _get_conv_fn(source, target):
    """Look up SO3.conversions.from_<source>_to_<target>."""
    return getattr(SO3.conversions, f"from_{source}_to_{target}")


# ---------------------------------------------------------------------------
# 1. Equivalence: wrapper == manual two-step
# ---------------------------------------------------------------------------


@pytest.mark.parametrize("pair", _PAIRS, ids=[f"{s}->{t}" for s, t in _PAIRS])
@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_equivalence(pair, backend, batch_dims, pass_xp):
    source, target = pair
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    x = _make_input(source, batch_dims, backend)

    fn = _get_conv_fn(source, target)
    result = fn(x, **xp_kwargs)
    expected = _manual_convert(x, source, target)

    result_np = np.array(result)
    expected_np = np.array(expected)

    atol = ATOL[32]
    if target in ("quat_wxyz", "quat_xyzw"):
        # quaternion sign ambiguity
        dots = np.sum(result_np * expected_np, axis=-1)
        assert np.allclose(np.abs(dots), 1.0, atol=atol), f"{source}->{target} quat mismatch"
    else:
        assert np.allclose(result_np, expected_np, atol=atol), f"{source}->{target} mismatch"


# ---------------------------------------------------------------------------
# 2. Round-trip: B_to_A(A_to_B(x)) ≈ x
# ---------------------------------------------------------------------------

_ROUND_TRIP_PAIRS = [
    ("axis_angle", "matrix"),
    ("axis_angle", "quat_wxyz"),
    ("axis_angle", "sixd"),
    ("matrix", "quat_wxyz"),
    ("matrix", "sixd"),
    ("quat_wxyz", "quat_xyzw"),
    ("quat_wxyz", "sixd"),
]


@pytest.mark.parametrize("pair", _ROUND_TRIP_PAIRS, ids=[f"{a}<->{b}" for a, b in _ROUND_TRIP_PAIRS])
@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_round_trip(pair, backend, batch_dims):
    a, b = pair
    x = _make_input(a, batch_dims, backend)

    a_to_b = _get_conv_fn(a, b)
    b_to_a = _get_conv_fn(b, a)

    recovered = b_to_a(a_to_b(x))

    x_np = np.array(x)
    rec_np = np.array(recovered)

    atol = ATOL[32]
    if a in ("quat_wxyz", "quat_xyzw"):
        dots = np.sum(x_np * rec_np, axis=-1)
        assert np.allclose(np.abs(dots), 1.0, atol=atol), f"round-trip {a}<->{b} failed"
    else:
        assert np.allclose(x_np, rec_np, atol=atol), f"round-trip {a}<->{b} failed"


# ---------------------------------------------------------------------------
# 3. Euler convention threading
# ---------------------------------------------------------------------------


@pytest.mark.parametrize("convention", _EULER_CONVENTIONS)
@pytest.mark.parametrize("backend", TEST_BACKENDS)
def test_euler_convention_threading(convention, backend):
    q = random_quaternion(batch_dims=(5,), backend=backend)

    # euler as source
    euler = SO3.to_euler(q, convention=convention)
    result = SO3.conversions.from_euler_to_matrix(euler, convention=convention)
    expected = _manual_convert(euler, "euler", "matrix", convention=convention)
    assert np.allclose(np.array(result), np.array(expected), atol=ATOL[32])

    # euler as target
    aa = SO3.to_axis_angle(q)
    result2 = SO3.conversions.from_axis_angle_to_euler(aa, convention=convention)
    expected2 = _manual_convert(aa, "axis_angle", "euler", convention=convention)
    assert np.allclose(np.array(result2), np.array(expected2), atol=ATOL[32])



import numpy as np
import pytest
from conftest import ATOL, TEST_BACKENDS, TEST_BATCH_DIMS, TEST_PASS_XP, TEST_PRECISIONS, get_xp_kwargs, random_quaternion

from nanomanifold import SO3


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_exact_pi_rotation_all_axes(backend, precision, pass_xp):
    """Test logarithm at exactly π rotation for all principal axes."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)

    # Exactly π rotations around principal axes
    # q = cos(θ/2) + sin(θ/2) * (xi + yj + zk)
    # For θ = π: cos(π/2) = 0, sin(π/2) = 1

    pi_rotations = [
        xp.asarray([0.0, 1.0, 0.0, 0.0]),  # π around x-axis
        xp.asarray([0.0, 0.0, 1.0, 0.0]),  # π around y-axis
        xp.asarray([0.0, 0.0, 0.0, 1.0]),  # π around z-axis
    ]

    expected_logs = [
        [np.pi, 0.0, 0.0],
        [0.0, np.pi, 0.0],
        [0.0, 0.0, np.pi],
    ]

    for q, expected in zip(pi_rotations, expected_logs):
        log_result = SO3.log(q, **xp_kwargs)
        log_np = np.array(log_result)

        # The magnitude should be π
        magnitude = np.linalg.norm(log_np)
        assert np.allclose(magnitude, np.pi, atol=ATOL[precision])

        # The direction should match the expected axis
        direction = log_np / magnitude
        expected_direction = np.array(expected) / np.pi

        # Check direction (might be negated due to axis ambiguity)
        dot_product = np.dot(direction, expected_direction)
        assert np.allclose(np.abs(dot_product), 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_exact_pi_rotation_arbitrary_axes(backend, precision, pass_xp):
    """Test logarithm at exactly π rotation for arbitrary axes."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)

    # π rotations around arbitrary normalized axes
    axes = [
        [1.0, 1.0, 0.0],  # 45° between x and y
        [1.0, 0.0, 1.0],  # 45° between x and z
        [0.0, 1.0, 1.0],  # 45° between y and z
        [1.0, 1.0, 1.0],  # diagonal
        [2.0, -1.0, 3.0],  # arbitrary
    ]

    for axis in axes:
        # Normalize axis
        axis_np = np.array(axis)
        axis_np = axis_np / np.linalg.norm(axis_np)

        # Create quaternion for π rotation: q = [0, axis]
        q = xp.asarray([0.0, axis_np[0], axis_np[1], axis_np[2]])

        # Test log
        log_result = SO3.log(q, **xp_kwargs)
        log_np = np.array(log_result)

        # Magnitude should be π
        magnitude = np.linalg.norm(log_np)
        assert np.allclose(magnitude, np.pi, atol=ATOL[precision])

        # Test that exp(log(q)) = q (up to sign)
        exp_result = SO3.exp(log_result, **xp_kwargs)
        exp_np = np.array(exp_result)

        q_np = np.array(q)
        dot_product = np.sum(q_np * exp_np)
        assert np.allclose(np.abs(dot_product), 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_near_pi_rotation_stability(backend, precision, pass_xp):
    """Test numerical stability as rotation approaches π from different directions."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)

    # Test angles approaching π
    epsilons = [1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6, 1e-7, 1e-8]

    for eps in epsilons:
        # Angles just below and above π
        angles = [np.pi - eps, np.pi + eps]

        for angle in angles:
            # Rotation around z-axis using from_axis_angle
            axis_angle = xp.asarray([0.0, 0.0, angle])
            q = SO3.from_axis_angle(axis_angle, **xp_kwargs)

            # Test log
            log_result = SO3.log(q, **xp_kwargs)

            # Test exp(log(q)) = q
            exp_result = SO3.exp(log_result, **xp_kwargs)
            exp_np = np.array(exp_result)

            q_np = np.array(q)
            dot_product = np.sum(q_np * exp_np)

            # Tolerance should degrade gracefully as we approach singularity
            tol = max(ATOL[precision], eps * 10)
            assert np.allclose(np.abs(dot_product), 1.0, atol=tol)


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_quaternion_double_cover_mean(backend, batch_dims, precision, pass_xp):
    """Test mean computation with quaternion double cover (q and -q)."""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    # Create random quaternions
    q1 = random_quaternion(batch_dims, backend, precision)
    q2 = random_quaternion(batch_dims, backend, precision)
    q3 = random_quaternion(batch_dims, backend, precision)

    # Test different sign combinations
    sign_combinations = [
        [q1, q2, q3],
        [q1, q2, -q3],
        [q1, -q2, q3],
        [q1, -q2, -q3],
        [-q1, q2, q3],
        [-q1, q2, -q3],
        [-q1, -q2, q3],
        [-q1, -q2, -q3],
    ]

    results = []
    for quats in sign_combinations:
        mean_result = SO3.mean(quats, **xp_kwargs)
        results.append(np.array(mean_result))

    # All means should represent the same rotation (up to sign)
    reference = results[0]
    for i in range(1, len(results)):
        dot_product = np.sum(reference * results[i], axis=-1)
        assert np.allclose(np.abs(dot_product), 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_slerp_exact_antipodal(backend, precision, pass_xp):
    """Test SLERP with exactly antipodal quaternions."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)

    # Test multiple antipodal pairs
    pairs = [
        (xp.asarray([1.0, 0.0, 0.0, 0.0]), xp.asarray([-1.0, 0.0, 0.0, 0.0])),  # identity and -identity
        (xp.asarray([0.0, 1.0, 0.0, 0.0]), xp.asarray([0.0, -1.0, 0.0, 0.0])),  # π around x and -π around x
        (xp.asarray([0.5, 0.5, 0.5, 0.5]), xp.asarray([-0.5, -0.5, -0.5, -0.5])),  # normalized diagonal
    ]

    t_values = xp.asarray([0.0, 0.25, 0.5, 0.75, 1.0])

    for q1, q2 in pairs:
        result = SO3.slerp(q1, q2, t_values, **xp_kwargs)
        result_np = np.array(result)

        # All results should be unit quaternions
        norms = np.linalg.norm(result_np, axis=-1)
        assert np.allclose(norms, 1.0, atol=ATOL[precision])

        # Check endpoints
        assert np.allclose(np.abs(np.sum(result_np[0] * np.array(q1))), 1.0, atol=ATOL[precision])
        assert np.allclose(np.abs(np.sum(result_np[-1] * np.array(q2))), 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_log_exp_at_branch_cut(backend, precision, pass_xp):
    """Test log/exp behavior at the branch cut (rotations by exactly π)."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)

    # Create rotations at different positions on the branch cut
    # These are π rotations around axes in the xy-plane
    num_angles = 8
    for i in range(num_angles):
        angle = 2 * np.pi * i / num_angles
        axis = np.array([np.cos(angle), np.sin(angle), 0.0])

        # Create π rotation quaternion: q = [0, axis]
        q = xp.asarray([0.0, axis[0], axis[1], axis[2]])

        # Log should give magnitude π
        log_result = SO3.log(q, **xp_kwargs)
        log_np = np.array(log_result)
        magnitude = np.linalg.norm(log_np)
        assert np.allclose(magnitude, np.pi, atol=ATOL[precision])

        # Multiple applications of log/exp should be stable
        current = q
        for _ in range(5):
            current = SO3.exp(SO3.log(current, **xp_kwargs), **xp_kwargs)

        current_np = np.array(current)
        q_np = np.array(q)
        dot_product = np.sum(current_np * q_np)
        assert np.allclose(np.abs(dot_product), 1.0, atol=ATOL[precision] * 10)


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_gimbal_lock_euler_conversions(backend, pass_xp):
    """Test Euler angle conversions at gimbal lock configurations."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)

    # Gimbal lock occurs when the middle rotation is ±90°
    # For ZYX convention, this is when pitch = ±π/2

    # Create quaternions that result in gimbal lock
    gimbal_quaternions = []

    # Pitch = π/2 (looking straight up)
    q_up = SO3.from_euler(xp.asarray([[0.0, np.pi / 2, 0.0]]), convention="ZYX", **xp_kwargs)
    gimbal_quaternions.append(q_up)

    # Pitch = -π/2 (looking straight down)
    q_down = SO3.from_euler(xp.asarray([[0.0, -np.pi / 2, 0.0]]), convention="ZYX", **xp_kwargs)
    gimbal_quaternions.append(q_down)

    for q in gimbal_quaternions:
        # Convert to Euler angles and back
        euler = SO3.to_euler(q, convention="ZYX", **xp_kwargs)
        q_reconstructed = SO3.from_euler(euler, convention="ZYX", **xp_kwargs)

        # Check that we get the same rotation (up to sign)
        q_np = np.array(q).reshape(-1, 4)
        q_reconstructed_np = np.array(q_reconstructed).reshape(-1, 4)

        for i in range(len(q_np)):
            dot_product = np.sum(q_np[i] * q_reconstructed_np[i])
            assert np.allclose(np.abs(dot_product), 1.0, atol=1e-6)


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_zero_rotation_stability(backend, precision, pass_xp):
    """Test stability near zero rotation (identity)."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)

    # Very small rotation angles
    small_angles = [1e-8, 1e-10, 1e-12, 1e-14, 1e-15]

    for angle in small_angles:
        # Small rotation around random axis
        axis = np.random.randn(3)
        axis = axis / np.linalg.norm(axis)

        # Create tangent vector
        tangent = xp.asarray(angle * axis)

        # Test exp
        q = SO3.exp(tangent, **xp_kwargs)
        q_np = np.array(q)

        # Should be very close to identity
        identity = np.array([1.0, 0.0, 0.0, 0.0])
        assert np.allclose(q_np, identity, atol=angle * 2)

        # Test log(exp(v)) = v
        log_result = SO3.log(q, **xp_kwargs)
        log_np = np.array(log_result)
        tangent_np = np.array(tangent)

        # For very small angles, numerical precision limits accuracy
        tol = max(ATOL[precision], angle * 100)
        assert np.allclose(log_np, tangent_np, atol=tol)


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_log_matches_canonicalized_quaternion_near_two_pi(backend, precision, pass_xp):
    """Log should match canonical quaternion representation near 2π rotations."""

    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)

    np_dtype = getattr(np, f"float{precision}")

    axes = [
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [0.0, 0.0, 1.0],
        [1.0, -2.0, 0.5],
    ]

    deltas = [1e-2, 5e-3, 1e-3]

    for axis in axes:
        axis_np = np.array(axis, dtype=np.float64)
        axis_np = axis_np / np.linalg.norm(axis_np)

        for delta in deltas:
            theta = 2.0 * np.pi - delta
            half_angle = theta / 2.0

            q_np = np.concatenate(
                (
                    [np.cos(half_angle)],
                    np.sin(half_angle) * axis_np,
                )
            ).astype(np_dtype)

            q = xp.asarray(q_np)

            log_direct = np.array(SO3.log(q, **xp_kwargs))

            canonical_q = SO3.canonicalize(q, **xp_kwargs)
            log_canonical = np.array(SO3.log(canonical_q, **xp_kwargs))

            assert np.allclose(
                log_direct,
                log_canonical,
                atol=max(ATOL[precision], delta * 1e-3),
            )


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_log_direction_near_two_pi_rotation(backend, precision, pass_xp):
    """Log should pick the minimal axis-angle direction for rotations near 2π."""

    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)

    np_dtype = getattr(np, f"float{precision}")

    axes = [
        [0.0, 0.0, 1.0],
        [1.0, 0.0, 0.0],
        [0.3, -0.5, 0.2],
    ]

    deltas = [1e-2, 5e-3, 1e-3]

    for axis in axes:
        axis_np = np.array(axis, dtype=np.float64)
        axis_np = axis_np / np.linalg.norm(axis_np)

        for delta in deltas:
            theta = 2.0 * np.pi - delta
            half_angle = theta / 2.0

            q_np = np.concatenate(
                (
                    [np.cos(half_angle)],
                    np.sin(half_angle) * axis_np,
                )
            ).astype(np_dtype)

            q = xp.asarray(q_np)

            log_result = np.array(SO3.log(q, **xp_kwargs))

            expected = -delta * axis_np
            tol = max(ATOL[precision] * 10, delta * 1e-3)
            assert np.allclose(log_result, expected, atol=tol)



import numpy as np

from nanomanifold.common import get_namespace_by_name

TEST_BACKENDS = ["numpy", "torch", "jax"]
TEST_PRECISIONS = [16, 32]
TEST_BATCH_DIMS = [
    (),
    (1,),
    (5,),
    (2, 3),
    (2, 3, 4, 1, 2),
]
TEST_PASS_XP = [False, True]

ATOL = {
    16: 2e-2,
    32: 2e-6,
    64: 1e-12,
}


def get_xp_kwargs(backend, pass_xp):
    """Return xp keyword arguments based on whether to pass xp explicitly."""
    if not pass_xp:
        return {}
    return {"xp": get_namespace_by_name(backend)}


np.random.seed(0)


def random_quaternion(batch_dims=(), backend="numpy", precision=32):
    """Generate random quaternion in SO(3) for testing.

    Args:
        batch_dims: Tuple of batch dimensions
        backend: Backend to use ("numpy", "torch", "jax")
        precision: Precision (16, 32, 64)

    Returns:
        Quaternion array of shape (*batch_dims, 4) as [w, x, y, z]
    """
    # Generate in numpy first
    shape = batch_dims + (4,)
    q_np = np.random.randn(*shape).astype(f"float{precision}")

    # Normalize to get unit quaternion
    norm = np.linalg.norm(q_np, axis=-1, keepdims=True)
    q_np = q_np / norm

    # Canonicalize to w >= 0
    mask = q_np[..., 0:1] < 0
    q_np = np.where(mask, -q_np, q_np)

    xp = get_namespace_by_name(backend)
    return xp.asarray(q_np)


def identity_quaternion(batch_dims=(), backend="numpy", precision=32):
    """Generate identity quaternion for testing.

    Args:
        batch_dims: Tuple of batch dimensions
        backend: Backend to use ("numpy", "torch", "jax")
        precision: Precision (16, 32, 64)

    Returns:
        Identity quaternion array of shape (*batch_dims, 4) as [1, 0, 0, 0]
    """
    # Create identity quaternion [1, 0, 0, 0] in numpy first
    shape = batch_dims + (4,)
    q_np = np.zeros(shape, dtype=f"float{precision}")
    q_np[..., 0] = 1  # w = 1, x = y = z = 0

    xp = get_namespace_by_name(backend)
    return xp.asarray(q_np)


def random_points(batch_dims=(), n_points=10, backend="numpy", precision=32):
    """Generate random 3D points for testing.

    Args:
        batch_dims: Tuple of batch dimensions
        n_points: Number of points to generate
        backend: Backend to use ("numpy", "torch", "jax")
        precision: Precision (16, 32, 64)

    Returns:
        Points array of shape (*batch_dims, n_points, 3)
    """
    # Generate in numpy first
    shape = batch_dims + (n_points, 3)
    points_np = np.random.randn(*shape).astype(f"float{precision}")

    xp = get_namespace_by_name(backend)
    return xp.asarray(points_np)


def random_se3(batch_dims=(), backend="numpy", precision=32):
    """Generate random SE(3) transformation for testing.

    Args:
        batch_dims: Tuple of batch dimensions
        backend: Backend to use ("numpy", "torch", "jax")
        precision: Precision (16, 32, 64)

    Returns:
        SE(3) array of shape (*batch_dims, 7) as [w, x, y, z, tx, ty, tz]
    """
    # Generate random quaternion
    quat = random_quaternion(batch_dims=batch_dims, backend=backend, precision=precision)

    # Generate random translation
    shape = batch_dims + (3,)
    translation_np = np.random.randn(*shape).astype(f"float{precision}")

    xp = get_namespace_by_name(backend)
    translation = xp.asarray(translation_np)

    # Combine quaternion and translation
    return xp.concatenate([quat, translation], axis=-1)


def identity_se3(batch_dims=(), backend="numpy", precision=32):
    """Generate identity SE(3) transformation for testing.

    Args:
        batch_dims: Tuple of batch dimensions
        backend: Backend to use ("numpy", "torch", "jax")
        precision: Precision (16, 32, 64)

    Returns:
        Identity SE(3) array of shape (*batch_dims, 7) as [1, 0, 0, 0, 0, 0, 0]
    """
    # Create identity quaternion and zero translation
    quat = identity_quaternion(batch_dims=batch_dims, backend=backend, precision=precision)

    shape = batch_dims + (3,)
    translation_np = np.zeros(shape, dtype=f"float{precision}")

    xp = get_namespace_by_name(backend)
    translation = xp.asarray(translation_np)

    # Combine quaternion and translation
    return xp.concatenate([quat, translation], axis=-1)



from types import ModuleType
from typing import Any

from jaxtyping import Float

from nanomanifold.common import get_namespace
from nanomanifold.SO3 import hat as so3_hat


def hat(v: Float[Any, "... 6"], *, xp: ModuleType | None = None) -> Float[Any, "... 4 4"]:
    """Map 6-vector to 4x4 matrix in se(3) (hat operator).

    The input 6-vector is [omega, rho] where:
    - omega (first 3) is the angular velocity
    - rho (last 3) is the translational velocity

    The output 4x4 matrix has the form:
        [[SO3.hat(omega), rho],
         [0,  0,  0,       0 ]]

    Args:
        v: (..., 6) array representing tangent vectors in se(3)
        xp: Array namespace (e.g. torch, jax.numpy). If None, auto-detected.

    Returns:
        (..., 4, 4) matrices in se(3)
    """
    if xp is None:
        xp = get_namespace(v)

    omega = v[..., :3]
    rho = v[..., 3:6]

    omega_hat = so3_hat(omega, xp=xp)  # (..., 3, 3)

    rho_col = rho[..., None]  # (..., 3, 1)
    top_block = xp.concatenate([omega_hat, rho_col], axis=-1)  # (..., 3, 4)

    zeros = xp.zeros_like(top_block[..., :1, :])  # (..., 1, 4)
    result = xp.concatenate([top_block, zeros], axis=-2)  # (..., 4, 4)

    return result



import pytest

from nanomanifold import SE3, SO3


def test_so3_from_matrix_gradient_finite_torch_at_identity():
    torch = pytest.importorskip("torch")
    matrix = torch.eye(3, dtype=torch.float64, requires_grad=True)

    quat = SO3.from_matrix(matrix)
    quat.sum().backward()

    assert torch.isfinite(matrix.grad).all()


def test_so3_from_matrix_gradient_finite_torch_near_identity():
    torch = pytest.importorskip("torch")
    matrix = torch.eye(3, dtype=torch.float64)
    matrix[0, 1] = 1e-12
    matrix[1, 0] = -1e-12
    matrix.requires_grad_(True)

    quat = SO3.from_matrix(matrix)
    quat.sum().backward()

    assert torch.isfinite(matrix.grad).all()


def test_se3_from_matrix_gradient_finite_torch_at_identity():
    torch = pytest.importorskip("torch")
    matrix = torch.eye(4, dtype=torch.float64, requires_grad=True)

    se3 = SE3.from_matrix(matrix)
    se3.sum().backward()

    assert torch.isfinite(matrix.grad).all()



import numpy as np
import pytest
from conftest import ATOL, TEST_BACKENDS, TEST_BATCH_DIMS, TEST_PASS_XP, TEST_PRECISIONS, get_xp_kwargs, random_quaternion
from scipy.spatial.transform import Rotation as R

from nanomanifold import SO3


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_axis_angle_conversion_cycle(backend, batch_dims, precision, pass_xp):
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    # Create a random SO3 quaternion
    quat = random_quaternion(batch_dims=batch_dims, backend=backend, precision=precision)

    # Convert to axis-angle representation
    axis_angle = SO3.to_axis_angle(quat, **xp_kwargs)

    assert axis_angle.dtype == quat.dtype
    assert axis_angle.shape[:-1] == quat.shape[:-1]

    # Convert back to quaternion
    quat_converted = SO3.from_axis_angle(axis_angle, **xp_kwargs)

    assert quat_converted.dtype == quat.dtype
    assert quat_converted.shape == quat.shape

    # Convert to numpy arrays and compare
    quat_np = np.array(quat)
    quat_converted_np = np.array(quat_converted)

    # Check quaternion equivalence (q and -q represent the same rotation)
    dot_products = np.sum(quat_np * quat_converted_np, axis=-1)
    assert np.allclose(np.abs(dot_products), 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_axis_angle_conversion_scipy(backend, batch_dims, pass_xp):
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    # Create a random SO3 quaternion
    quat = random_quaternion(batch_dims=batch_dims, backend=backend, precision=32)

    # Convert to axis-angle representation using nanomanifold
    axis_angle = SO3.to_axis_angle(quat, **xp_kwargs)

    # Convert to axis-angle representation using scipy
    quat_np = np.array(quat)
    # Convert from [w, x, y, z] to scipy's [x, y, z, w] format
    quat_scipy = np.concatenate([quat_np[..., 1:4], quat_np[..., 0:1]], axis=-1)
    r = R.from_quat(quat_scipy.reshape(-1, 4))
    axis_angle_scipy = r.as_rotvec().reshape(axis_angle.shape)

    assert axis_angle.dtype == quat.dtype
    assert axis_angle.shape == axis_angle_scipy.shape

    axis_angle_np = np.array(axis_angle)
    assert np.allclose(axis_angle_np, axis_angle_scipy, atol=1e-6)


@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_axis_angle_differentiality_torch(batch_dims):
    torch = pytest.importorskip("torch")
    # Use double precision for gradient checking as recommended by PyTorch
    dtype = torch.float64

    # Random quaternion input
    quat = random_quaternion(batch_dims=batch_dims, backend="torch").to(dtype).requires_grad_(True)

    # Check gradients of SO3.to_axis_angle
    def f(q):
        return SO3.to_axis_angle(q)

    assert torch.autograd.gradcheck(f, (quat,), eps=1e-6, atol=1e-5)

    # Axis-angle input from random quaternion
    axis_angle = SO3.to_axis_angle(quat.detach()).requires_grad_(True)

    # Check gradients of SO3.from_axis_angle
    def g(aa):
        return SO3.from_axis_angle(aa)

    assert torch.autograd.gradcheck(g, (axis_angle,), eps=1e-6, atol=1e-5)


@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_axis_angle_differentiality_torch_at_singularity(batch_dims):
    torch = pytest.importorskip("torch")
    dtype = torch.float64

    axis_angle = torch.zeros(batch_dims + (3,), dtype=dtype)
    axis_angle[..., 0] = np.pi

    quat = SO3.from_axis_angle(axis_angle).requires_grad_(True)

    axis_angle_out = SO3.to_axis_angle(quat)
    axis_angle_out.sum().backward()
    assert torch.isfinite(quat.grad).all()

    axis_angle = axis_angle.clone().requires_grad_(True)
    quat_out = SO3.from_axis_angle(axis_angle)
    quat_out.sum().backward()
    assert torch.isfinite(axis_angle.grad).all()



import math
from types import ModuleType
from typing import Any

from jaxtyping import Float

from nanomanifold import common
from nanomanifold.common import get_namespace
from nanomanifold.SO3 import exp as so3_exp
from nanomanifold.SO3 import hat


def exp(tangent_vector: Float[Any, "... 6"], *, xp: ModuleType | None = None) -> Float[Any, "... 7"]:
    """Compute the exponential map from se(3) tangent space to SE(3) manifold.

    The exponential map takes a tangent vector in the Lie algebra se(3)
    and returns the corresponding SE(3) transformation. This is the inverse
    operation of log().

    The se(3) exponential map takes a 6-vector [ω, ρ] where:
    - ω ∈ ℝ³ is the angular velocity (rotation part)
    - ρ ∈ ℝ³ is the translational velocity

    The formula involves:
    - R = exp_SO3(ω) for the rotation quaternion
    - t = V * ρ where V is the left Jacobian matrix

    Args:
        tangent_vector: Tangent vector in se(3) as [ω, ρ] of shape (..., 6)
        xp: Array namespace (e.g. torch, jax.numpy). If None, auto-detected.

    Returns:
        SE(3) transformation in [w, x, y, z, tx, ty, tz] format of shape (..., 7)
    """
    if xp is None:
        xp = get_namespace(tangent_vector)

    omega = tangent_vector[..., :3]
    rho = tangent_vector[..., 3:6]

    q = so3_exp(omega, xp=xp)
    omega_norm = xp.linalg.norm(omega, axis=-1, keepdims=True)

    thresh = xp.asarray(math.sqrt(common.safe_eps(omega.dtype, xp, scale=1.0)), dtype=omega.dtype)
    small_angle_mask = omega_norm < thresh

    omega_cross = hat(omega, xp=xp)
    omega_cross_sq = xp.matmul(omega_cross, omega_cross)

    identity = common.eye_as(omega, batch_dims=omega.shape[:-1], xp=xp)

    V_small = identity + 0.5 * omega_cross + (1.0 / 12.0) * omega_cross_sq

    cos_norm = xp.cos(omega_norm)
    sin_norm = xp.sin(omega_norm)

    safe_norm = xp.where(small_angle_mask, xp.ones_like(omega_norm), omega_norm)
    safe_norm_sq = xp.where(small_angle_mask, xp.ones_like(omega_norm), omega_norm**2)
    safe_norm_cub = safe_norm_sq * safe_norm

    A = (1.0 - cos_norm) / safe_norm_sq
    B = (safe_norm - sin_norm) / safe_norm_cub
    A = xp.reshape(A, A.shape[:-1])[..., None, None]
    B = xp.reshape(B, B.shape[:-1])[..., None, None]

    V_large = identity + A * omega_cross + B * omega_cross_sq

    mask = xp.reshape(small_angle_mask, small_angle_mask.shape[:-1])[..., None, None]
    V = xp.where(mask, V_small, V_large)
    V = xp.reshape(V, omega.shape[:-1] + (3, 3))

    t = xp.matmul(V, rho[..., None])[..., 0]

    se3 = xp.concatenate([q, t], axis=-1)

    return se3



from importlib.metadata import PackageNotFoundError, version

from . import SE3, SO3

try:
    __version__ = version("nanomanifold")
except PackageNotFoundError:
    __version__ = "0+unknown"

__all__ = ["SO3", "SE3", "__version__"]



from types import ModuleType
from typing import Any

from jaxtyping import Float

from nanomanifold import common
from nanomanifold.common import get_namespace


def distance(q1: Float[Any, "... 4"], q2: Float[Any, "... 4"], *, xp: ModuleType | None = None) -> Float[Any, "..."]:
    """Compute the angular distance between two quaternions representing SO(3) rotations.

    The angular distance is the smallest angle needed to rotate from one orientation
    to another, measured in radians. This is equivalent to the geodesic distance
    on the SO(3) manifold.

    Args:
        q1: First quaternion in [w, x, y, z] format
        q2: Second quaternion in [w, x, y, z] format
        xp: Array namespace (e.g. torch, jax.numpy). If None, auto-detected.

    Returns:
        Angular distance in radians, in range [0, π]
    """
    if xp is None:
        xp = get_namespace(q1)

    eps = common.safe_eps(q1.dtype, xp)
    eps_arr = xp.asarray(eps, dtype=q1.dtype)
    eps_sq = eps_arr * eps_arr

    norm1 = xp.linalg.norm(q1, axis=-1, keepdims=True)
    norm2 = xp.linalg.norm(q2, axis=-1, keepdims=True)
    q1_unit = q1 / xp.maximum(norm1, eps_arr)
    q2_unit = q2 / xp.maximum(norm2, eps_arr)

    w1 = q1_unit[..., :1]
    v1 = q1_unit[..., 1:]
    w2 = q2_unit[..., :1]
    v2 = q2_unit[..., 1:]

    cross = xp.stack(
        [
            v1[..., 1] * v2[..., 2] - v1[..., 2] * v2[..., 1],
            v1[..., 2] * v2[..., 0] - v1[..., 0] * v2[..., 2],
            v1[..., 0] * v2[..., 1] - v1[..., 1] * v2[..., 0],
        ],
        axis=-1,
    )

    vec = w1 * v2 - w2 * v1 - cross
    vec_sq = xp.sum(vec * vec, axis=-1)
    vec_norm = xp.sqrt(xp.maximum(vec_sq, eps_sq))
    w = w1 * w2 + xp.sum(v1 * v2, axis=-1, keepdims=True)
    w_abs = xp.abs(w[..., 0])
    angle = 2.0 * xp.atan2(vec_norm, w_abs)

    return xp.where(vec_sq <= eps_sq, xp.zeros_like(angle), angle)



from types import ModuleType
from typing import Any

from jaxtyping import Float

from nanomanifold.common import get_namespace

from ..canonicalize import canonicalize as canonicalize_se3


def from_rt(quat: Float[Any, "... 4"], translation: Float[Any, "... 3"], *, xp: ModuleType | None = None) -> Float[Any, "... 7"]:
    """Create SE(3) representation from rotation quaternion and translation.

    Args:
        quat: Rotation quaternion (..., 4) as [w, x, y, z]
        translation: Translation vector (..., 3)
        xp: Array namespace (e.g. torch, jax.numpy). If None, auto-detected.

    Returns:
        SE(3) representation (..., 7) as [w, x, y, z, tx, ty, tz] with the
        quaternion canonicalized to have a non-negative scalar component.
    """
    if xp is None:
        xp = get_namespace(quat)
    se3 = xp.concatenate([quat, translation], axis=-1)
    return canonicalize_se3(se3, xp=xp)


def to_rt(se3: Float[Any, "... 7"]) -> tuple[Float[Any, "... 4"], Float[Any, "... 3"]]:
    """Extract rotation quaternion and translation from SE(3) representation.

    Args:
        se3: SE(3) representation (..., 7) as [w, x, y, z, tx, ty, tz]

    Returns:
        quat: Rotation quaternion (..., 4) as [w, x, y, z]
        translation: Translation vector (..., 3)
    """
    quat = se3[..., :4]
    translation = se3[..., 4:7]
    return quat, translation



import numpy as np
import pytest
from conftest import ATOL, TEST_BACKENDS, TEST_BATCH_DIMS, TEST_PASS_XP, TEST_PRECISIONS, get_xp_kwargs, random_quaternion
from scipy.spatial.transform import Rotation as R

from nanomanifold import SO3


def get_dtype(backend, precision):
    """Get appropriate dtype for given backend and precision."""
    if backend == "numpy":
        return getattr(np, f"float{precision}")
    elif backend == "torch":
        torch = pytest.importorskip("torch")
        return getattr(torch, f"float{precision}")
    else:  # jax
        jax = pytest.importorskip("jax")
        return getattr(jax.numpy, f"float{precision}")


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_mean_single_quaternion(backend, batch_dims, precision, pass_xp):
    """Test that mean of a single quaternion returns that quaternion"""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    q = random_quaternion(batch_dims=batch_dims, backend=backend, precision=precision)

    result = SO3.mean([q], **xp_kwargs)

    assert result.dtype == q.dtype
    assert result.shape == q.shape

    q_np = np.array(q)
    result_np = np.array(result)

    # Check quaternion equivalence (q and -q represent same rotation)
    dot_product = np.sum(q_np * result_np, axis=-1)
    assert np.allclose(np.abs(dot_product), 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_mean_identical_quaternions(backend, batch_dims, precision, pass_xp):
    """Test that mean of identical quaternions returns that quaternion"""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    q = random_quaternion(batch_dims=batch_dims, backend=backend, precision=precision)

    # Create list of identical quaternions
    quaternions = [q, q, q, q]
    result = SO3.mean(quaternions, **xp_kwargs)

    assert result.dtype == q.dtype
    assert result.shape == q.shape

    q_np = np.array(q)
    result_np = np.array(result)

    # Check quaternion equivalence
    dot_product = np.sum(q_np * result_np, axis=-1)
    assert np.allclose(np.abs(dot_product), 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_mean_antipodal_quaternions(backend, batch_dims, precision, pass_xp):
    """Test mean of antipodal quaternions (q and -q)"""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    q = random_quaternion(batch_dims=batch_dims, backend=backend, precision=precision)

    # Get appropriate backend namespace
    common = pytest.importorskip("nanomanifold.common")
    common.get_namespace_by_name(backend)

    q_neg = -q
    result = SO3.mean([q, q_neg], **xp_kwargs)

    assert result.dtype == q.dtype
    assert result.shape == q.shape

    q_np = np.array(q)
    result_np = np.array(result)

    # Result should be equivalent to either q or -q
    dot_product = np.sum(q_np * result_np, axis=-1)
    assert np.allclose(np.abs(dot_product), 1.0, atol=ATOL[precision])

    # Result should be unit quaternion
    norm = np.linalg.norm(result_np, axis=-1)
    assert np.allclose(norm, 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_weighted_mean_uniform_weights(backend, batch_dims, precision, pass_xp):
    """Test that weighted_mean with uniform weights equals mean"""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    quaternions = [random_quaternion(batch_dims=batch_dims, backend=backend, precision=precision) for _ in range(4)]

    # Get appropriate backend namespace
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend)
    dtype = quaternions[0].dtype

    # Uniform weights as array
    weights = xp.ones(batch_dims + (4,), dtype=dtype)

    result_weighted = SO3.weighted_mean(quaternions, weights, **xp_kwargs)
    result_mean = SO3.mean(quaternions, **xp_kwargs)

    assert result_weighted.dtype == result_mean.dtype
    assert result_weighted.shape == result_mean.shape

    result_weighted_np = np.array(result_weighted)
    result_mean_np = np.array(result_mean)

    # Check quaternion equivalence
    dot_product = np.sum(result_weighted_np * result_mean_np, axis=-1)
    assert np.allclose(np.abs(dot_product), 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_weighted_mean_single_nonzero_weight(backend, batch_dims, precision, pass_xp):
    """Test weighted_mean with only one nonzero weight"""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    quaternions = [random_quaternion(batch_dims=batch_dims, backend=backend, precision=precision) for _ in range(4)]

    # Get appropriate backend namespace
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend)
    dtype = quaternions[0].dtype

    # Only second quaternion has nonzero weight
    weights = xp.asarray([0.0, 1.0, 0.0, 0.0], dtype=dtype)
    # Broadcast to proper batch shape
    if batch_dims:
        weights = xp.broadcast_to(weights, batch_dims + (4,))

    result = SO3.weighted_mean(quaternions, weights, **xp_kwargs)

    assert result.dtype == quaternions[1].dtype
    assert result.shape == quaternions[1].shape

    q1_np = np.array(quaternions[1])
    result_np = np.array(result)

    # Result should be equivalent to the second quaternion
    dot_product = np.sum(q1_np * result_np, axis=-1)
    assert np.allclose(np.abs(dot_product), 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_weighted_mean_array_weights(backend, precision, pass_xp):
    """Test weighted_mean with array weights for batched inputs"""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    batch_dims = (2, 3)

    # Get appropriate backend namespace
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend)
    dtype = get_dtype(backend, precision)

    quaternions = [random_quaternion(batch_dims=batch_dims, backend=backend, precision=precision) for _ in range(3)]

    # Create array weights with shape batch_dims + (n_quats,) = (2, 3, 3)
    weights = xp.asarray(
        [
            [[1.0, 2.0, 1.0], [0.5, 1.0, 0.5], [2.0, 0.5, 1.5]],  # batch [0, :, :] weights for 3 quats
            [[2.0, 1.0, 1.0], [1.0, 0.5, 0.5], [1.5, 0.8, 1.2]],  # batch [1, :, :] weights for 3 quats
        ],
        dtype=dtype,
    )

    result = SO3.weighted_mean(quaternions, weights, **xp_kwargs)

    assert result.dtype == quaternions[0].dtype
    assert result.shape == batch_dims + (4,)

    result_np = np.array(result)

    # Result should be unit quaternions
    norms = np.linalg.norm(result_np, axis=-1)
    assert np.allclose(norms, 1.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_mean_scipy_comparison(backend, pass_xp):
    """Compare mean results with scipy implementation for special cases"""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)

    # Get appropriate backend namespace
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend)

    # Test with known rotations
    sqrt2_inv = 1.0 / np.sqrt(2.0)

    # Create quaternions in the appropriate backend
    identity = xp.asarray([1.0, 0.0, 0.0, 0.0])
    q_x90 = xp.asarray([sqrt2_inv, sqrt2_inv, 0.0, 0.0])
    q_y90 = xp.asarray([sqrt2_inv, 0.0, sqrt2_inv, 0.0])
    q_z90 = xp.asarray([sqrt2_inv, 0.0, 0.0, sqrt2_inv])

    quaternions = [identity, q_x90, q_y90, q_z90]
    result = SO3.mean(quaternions, **xp_kwargs)

    # Convert to numpy for scipy
    result_np = np.array(result)

    # Convert to scipy format [x, y, z, w] and compute mean
    quaternions_scipy = []
    for q in quaternions:
        q_np = np.array(q)
        q_scipy = np.concatenate([q_np[1:4], q_np[0:1]], axis=-1)  # [w,x,y,z] -> [x,y,z,w]
        quaternions_scipy.append(q_scipy)

    # Scipy's mean rotation
    rotations = R.from_quat(quaternions_scipy)
    mean_rotation = rotations.mean()
    scipy_result = mean_rotation.as_quat()

    # Convert back to [w, x, y, z] format
    scipy_result = np.concatenate([scipy_result[3:4], scipy_result[0:3]], axis=-1)

    # Check quaternion equivalence (allowing for sign flip)
    dot_product = np.sum(result_np * scipy_result)
    assert np.allclose(np.abs(dot_product), 1.0, atol=1e-6)


def test_weighted_mean_zero_norm_quaternion():
    """Test that weighted_mean handles zero norm quaternion gracefully"""
    zero_quat = np.array([0.0, 0.0, 0.0, 0.0])
    valid_quat = np.array([1.0, 0.0, 0.0, 0.0])
    weights = np.array([1.0, 1.0])

    # This should not crash but may give unexpected results due to normalization
    result = SO3.weighted_mean([zero_quat, valid_quat], weights)
    # The result should still be a unit quaternion due to normalization
    assert np.allclose(np.linalg.norm(result), 1.0, atol=1e-6)


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_mean_numerical_stability(backend, pass_xp):
    """Test numerical stability with very small rotations"""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    # Get appropriate backend namespace
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend)

    # Very small rotation angles
    identity = xp.asarray([1.0, 0.0, 0.0, 0.0])
    small_eps = 1e-8

    # Small rotations around different axes
    small_x = xp.asarray([1.0, small_eps, 0.0, 0.0])
    small_y = xp.asarray([1.0, 0.0, small_eps, 0.0])
    small_z = xp.asarray([1.0, 0.0, 0.0, small_eps])

    # Normalize
    small_x = small_x / xp.sqrt(xp.sum(small_x**2))
    small_y = small_y / xp.sqrt(xp.sum(small_y**2))
    small_z = small_z / xp.sqrt(xp.sum(small_z**2))

    quaternions = [identity, small_x, small_y, small_z]
    result = SO3.mean(quaternions, **xp_kwargs)

    # Result should be unit quaternion
    result_np = np.array(result)
    norm = np.linalg.norm(result_np)
    assert np.allclose(norm, 1.0, atol=1e-12)

    # Result should be close to identity for small rotations
    dot_with_identity = np.sum(result_np * np.array([1.0, 0.0, 0.0, 0.0]))
    assert dot_with_identity > 0.99


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_weighted_mean_consistency(backend, pass_xp):
    """Test that repeated weighted_mean calls are consistent"""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    # Get appropriate backend namespace
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend)

    quaternions = [random_quaternion((), backend, 32) for _ in range(5)]
    weights = xp.asarray([0.2, 0.3, 0.1, 0.25, 0.15], dtype=quaternions[0].dtype)

    # Compute mean multiple times
    results = []
    for _ in range(5):
        result = SO3.weighted_mean(quaternions, weights, **xp_kwargs)
        results.append(np.array(result))

    # All results should be identical (deterministic)
    for i in range(1, len(results)):
        dot_product = np.sum(results[0] * results[i])
        assert np.allclose(np.abs(dot_product), 1.0, atol=1e-12)


def test_weighted_mean_specific_case():
    """Test weighted mean for a specific known case"""
    # Identity and 90° rotation around z-axis
    identity = np.array([1.0, 0.0, 0.0, 0.0])
    sqrt2_inv = 1.0 / np.sqrt(2.0)
    q_z90 = np.array([sqrt2_inv, 0.0, 0.0, sqrt2_inv])

    # Equal weights should give result similar to slerp at t=0.5
    weights = np.array([1.0, 1.0])
    result = SO3.weighted_mean([identity, q_z90], weights)
    result_np = np.array(result)

    # Compare with slerp at midpoint
    t_mid = np.array([0.5])
    slerp_result = SO3.slerp(identity, q_z90, t_mid)
    slerp_result_np = np.array(slerp_result)[0]  # Remove the N=1 dimension

    # Should agree up to sign ambiguity of unit quaternions
    dot_product = np.sum(result_np * slerp_result_np)
    assert np.allclose(np.abs(dot_product), 1.0, atol=1e-6)


def test_mean_unit_quaternion_output():
    """Test that mean always produces unit quaternions"""
    # Random quaternions with various magnifications
    base_quats = [
        np.array([1.0, 0.5, -0.3, 0.8]),
        np.array([0.7, -0.2, 0.6, 0.1]),
        np.array([-0.4, 0.9, 0.1, -0.7]),
    ]

    # Normalize them properly
    quaternions = []
    for q in base_quats:
        q_normalized = q / np.linalg.norm(q)
        quaternions.append(q_normalized)

    result = SO3.mean(quaternions)
    result_np = np.array(result)

    # Must be unit quaternion
    norm = np.linalg.norm(result_np)
    assert np.allclose(norm, 1.0, atol=1e-12)


def test_weighted_mean_invariant_to_weight_scaling():
    """Test that weighted_mean is invariant to scaling all weights by same factor"""
    quaternions = [random_quaternion((), "numpy", 32) for _ in range(4)]

    weights1 = np.array([0.1, 0.3, 0.4, 0.2])
    weights2 = np.array([1.0, 3.0, 4.0, 2.0])  # Scaled by 10
    weights3 = np.array([0.05, 0.15, 0.2, 0.1])  # Scaled by 0.5

    result1 = SO3.weighted_mean(quaternions, weights1)
    result2 = SO3.weighted_mean(quaternions, weights2)
    result3 = SO3.weighted_mean(quaternions, weights3)

    result1_np = np.array(result1)
    result2_np = np.array(result2)
    result3_np = np.array(result3)

    # All should be equivalent
    dot12 = np.sum(result1_np * result2_np)
    dot13 = np.sum(result1_np * result3_np)

    assert np.allclose(np.abs(dot12), 1.0, atol=1e-12)
    assert np.allclose(np.abs(dot13), 1.0, atol=1e-12)


@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_mean_jittability_jax(batch_dims):
    """Test that mean functions are JIT-compatible with JAX"""
    jax = pytest.importorskip("jax")

    # Create test quaternions
    quaternions = [random_quaternion(batch_dims=batch_dims, backend="jax", precision=32) for _ in range(3)]

    # Create weights for weighted_mean test
    num_quats = len(quaternions)
    weights = jax.numpy.ones(batch_dims + (num_quats,), dtype=jax.numpy.float32)

    # Define JIT-compiled functions
    @jax.jit
    def jit_mean(quats):
        return SO3.mean(quats)

    @jax.jit
    def jit_weighted_mean(quats, w):
        return SO3.weighted_mean(quats, w)

    # Test that JIT compilation works
    result_mean = jit_mean(quaternions)
    result_weighted = jit_weighted_mean(quaternions, weights)

    # Verify results are valid quaternions
    assert jax.numpy.allclose(jax.numpy.linalg.norm(result_mean, axis=-1), 1.0, atol=1e-6)
    assert jax.numpy.allclose(jax.numpy.linalg.norm(result_weighted, axis=-1), 1.0, atol=1e-6)



from types import ModuleType
from typing import Any

from jaxtyping import Float

from .primitives.axis_angle import from_axis_angle


def exp(tangent_vector: Float[Any, "... 3"], *, xp: ModuleType | None = None) -> Float[Any, "... 4"]:
    """Compute the exponential map from so(3) tangent space to SO(3) manifold.

    The exponential map takes a tangent vector in the Lie algebra so(3)
    and returns the corresponding rotation quaternion. This is the inverse
    operation of log().

    The exponential map is mathematically equivalent to converting an axis-angle
    representation to its corresponding quaternion.

    Args:
        tangent_vector: Tangent vector in so(3) (axis-angle representation)
                       The magnitude is the rotation angle, direction is the rotation axis
        xp: Array namespace (e.g. torch, jax.numpy). If None, auto-detected.

    Returns:
        Quaternion in [w, x, y, z] format representing the rotation
    """
    return from_axis_angle(tangent_vector, xp=xp)



from types import ModuleType
from typing import Any

from jaxtyping import Float

from nanomanifold.common import get_namespace
from nanomanifold.SO3 import slerp as so3_slerp

from .canonicalize import canonicalize


def slerp(
    se3_1: Float[Any, "... 7"],
    se3_2: Float[Any, "... 7"],
    t: Float[Any, "... N"],
    *,
    xp: ModuleType | None = None,
) -> Float[Any, "... N 7"]:
    """Interpolate between two SE(3) transformations.

    Uses SO3.slerp for the rotation part and linear interpolation for the
    translation part. The interpolation parameter ``t`` follows the same
    semantics as SO3.slerp: output shape is ``(..., N, 7)``.

    Args:
        se3_1: Start SE(3) transformation in [w, x, y, z, tx, ty, tz] format.
        se3_2: End SE(3) transformation in [w, x, y, z, tx, ty, tz] format.
        t: Interpolation parameters. Last dimension N gives the number of
            interpolation points. t=0 returns se3_1, t=1 returns se3_2.
        xp: Array namespace (e.g. torch, jax.numpy). If None, auto-detected.

    Returns:
        Interpolated SE(3) transformations with shape (..., N, 7).
    """
    if xp is None:
        xp = get_namespace(se3_1)

    se3_1 = canonicalize(se3_1, xp=xp)
    se3_2 = canonicalize(se3_2, xp=xp)

    q1 = se3_1[..., :4]
    t1 = se3_1[..., 4:7]
    q2 = se3_2[..., :4]
    t2 = se3_2[..., 4:7]

    # Slerp for rotation: (..., N, 4)
    q_interp = so3_slerp(q1, q2, t, xp=xp)

    # Linear interpolation for translation: (..., N, 3)
    t_expanded = t[..., None]  # (..., N, 1)
    t1_expanded = t1[..., None, :]  # (..., 1, 3)
    t2_expanded = t2[..., None, :]  # (..., 1, 3)
    t_interp = (1.0 - t_expanded) * t1_expanded + t_expanded * t2_expanded

    return xp.concatenate([q_interp, t_interp], axis=-1)



from types import ModuleType
from typing import Any, Sequence

from jaxtyping import Float

from nanomanifold.common import get_namespace
from nanomanifold.SO3 import weighted_mean as so3_weighted_mean

from .canonicalize import canonicalize


def weighted_mean(
    transforms: Sequence[Float[Any, "... 7"]],
    weights: Float[Any, "... N"],
    *,
    xp: ModuleType | None = None,
) -> Float[Any, "... 7"]:
    """Compute the weighted mean of SE(3) transformations.

    Uses SO3.weighted_mean for the rotation part and a weighted average
    for the translation part.

    Args:
        transforms: Sequence of SE(3) transformations in [w, x, y, z, tx, ty, tz] format.
        weights: Array of weights with shape [..., N] where N is the number of transforms.
        xp: Array namespace (e.g. torch, jax.numpy). If None, auto-detected.

    Returns:
        Weighted mean SE(3) transformation with shape [..., 7].
    """
    if xp is None:
        xp = get_namespace(transforms[0])

    transforms_canon = [canonicalize(t, xp=xp) for t in transforms]
    quaternions = [t[..., :4] for t in transforms_canon]
    translations = [t[..., 4:7] for t in transforms_canon]

    # Weighted mean of rotations
    q_mean = so3_weighted_mean(quaternions, weights, xp=xp)

    # Weighted average of translations
    original_dtype = transforms[0].dtype
    weights_array = xp.asarray(weights, dtype=original_dtype)
    weights_normalized = weights_array / xp.sum(weights_array, axis=-1, keepdims=True)

    t_stack = xp.stack(translations, axis=-2)  # (..., N, 3)
    t_mean = xp.sum(weights_normalized[..., :, None] * t_stack, axis=-2)  # (..., 3)

    return xp.concatenate([q_mean, t_mean], axis=-1)


def mean(
    transforms: Sequence[Float[Any, "... 7"]],
    *,
    xp: ModuleType | None = None,
) -> Float[Any, "... 7"]:
    """Compute the mean of SE(3) transformations.

    Equivalent to weighted_mean with uniform weights.

    Args:
        transforms: Sequence of SE(3) transformations in [w, x, y, z, tx, ty, tz] format.
        xp: Array namespace (e.g. torch, jax.numpy). If None, auto-detected.

    Returns:
        Mean SE(3) transformation with shape [..., 7].
    """
    if len(transforms) == 0:
        raise ValueError("Cannot compute mean of empty transform sequence")

    if xp is None:
        xp = get_namespace(transforms[0])

    batch_shape = transforms[0].shape[:-1]
    num_transforms = len(transforms)
    weights = xp.broadcast_to(xp.ones_like(transforms[0][..., :1]), batch_shape + (num_transforms,))

    return weighted_mean(transforms, weights, xp=xp)



import numpy as np
import pytest
from conftest import ATOL, TEST_BACKENDS, TEST_BATCH_DIMS, TEST_PASS_XP, TEST_PRECISIONS, get_xp_kwargs, random_quaternion

from nanomanifold import SE3


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_rt_conversion_cycle(backend, batch_dims, precision, pass_xp):
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    # Create random quaternion and translation
    quat = random_quaternion(batch_dims=batch_dims, backend=backend, precision=precision)

    # Generate random translation
    from nanomanifold.common import get_namespace_by_name

    shape = batch_dims + (3,)
    translation_np = np.random.randn(*shape).astype(f"float{precision}")
    xp = get_namespace_by_name(backend)
    translation = xp.asarray(translation_np)

    # Convert to SE(3) representation
    se3 = SE3.from_rt(quat, translation, **xp_kwargs)

    assert se3.dtype == quat.dtype
    assert se3.shape[:-1] == quat.shape[:-1]
    assert se3.shape[-1] == 7

    # Convert back to rotation and translation
    quat_converted, translation_converted = SE3.to_rt(se3)

    assert quat_converted.dtype == quat.dtype
    assert quat_converted.shape == quat.shape
    assert translation_converted.dtype == translation.dtype
    assert translation_converted.shape == translation.shape

    # Convert to numpy arrays and compare
    quat_np = np.array(quat)
    quat_converted_np = np.array(quat_converted)
    translation_np_orig = np.array(translation)
    translation_converted_np = np.array(translation_converted)

    # Check quaternion equivalence (q and -q represent the same rotation)
    dot_products = np.sum(quat_np * quat_converted_np, axis=-1)
    assert np.allclose(np.abs(dot_products), 1.0, atol=ATOL[precision])

    # Check translation equivalence
    assert np.allclose(translation_np_orig, translation_converted_np, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_rt_matrix_consistency(backend, batch_dims, pass_xp):
    """Test that from_rt -> to_matrix -> from_matrix -> to_rt gives same result."""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    # Create random quaternion and translation
    quat = random_quaternion(batch_dims=batch_dims, backend=backend, precision=32)

    from nanomanifold.common import get_namespace_by_name

    shape = batch_dims + (3,)
    translation_np = np.random.randn(*shape).astype("float32")
    xp = get_namespace_by_name(backend)
    translation = xp.asarray(translation_np)

    # Route 1: from_rt -> to_matrix -> from_matrix -> to_rt
    se3_from_rt = SE3.from_rt(quat, translation, **xp_kwargs)
    matrix = SE3.to_matrix(se3_from_rt, **xp_kwargs)
    se3_from_matrix = SE3.from_matrix(matrix, **xp_kwargs)
    quat_final, translation_final = SE3.to_rt(se3_from_matrix)

    # Convert to numpy for comparison
    quat_np = np.array(quat)
    quat_final_np = np.array(quat_final)
    translation_np = np.array(translation)
    translation_final_np = np.array(translation_final)

    # Check quaternion equivalence (q and -q represent the same rotation)
    dot_products = np.sum(quat_np * quat_final_np, axis=-1)
    assert np.allclose(np.abs(dot_products), 1.0, atol=1e-6)

    # Check translation equivalence
    assert np.allclose(translation_np, translation_final_np, atol=1e-6)


@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_rt_differentiality_torch(batch_dims):
    torch = pytest.importorskip("torch")
    # Use double precision for gradient checking as recommended by PyTorch
    dtype = torch.float64

    # Random quaternion and translation inputs
    quat = random_quaternion(batch_dims=batch_dims, backend="torch").to(dtype).requires_grad_(True)
    translation = torch.randn(batch_dims + (3,), dtype=dtype, requires_grad=True)

    # Check gradients of SE3.from_rt
    def f(q, t):
        return SE3.from_rt(q, t)

    assert torch.autograd.gradcheck(f, (quat, translation), eps=1e-6, atol=1e-5)

    # SE3 input from random quaternion and translation
    se3 = SE3.from_rt(quat.detach(), translation.detach()).requires_grad_(True)

    # Check gradients of SE3.to_rt
    def g(s):
        q, t = SE3.to_rt(s)
        return q.sum() + t.sum()  # Need scalar output for gradcheck

    assert torch.autograd.gradcheck(g, (se3,), eps=1e-6, atol=1e-5)



import numpy as np
import pytest
from conftest import ATOL, TEST_BACKENDS, TEST_BATCH_DIMS, TEST_PASS_XP, TEST_PRECISIONS, get_xp_kwargs

from nanomanifold import SO3


def get_dtype(backend, precision):
    """Get dtype for given backend and precision.

    Args:
        backend: Backend name ("numpy", "torch", "jax")
        precision: Precision (16, 32, 64)

    Returns:
        Appropriate dtype for the backend
    """
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))

    if precision == 16:
        return xp.float16
    elif precision == 32:
        return xp.float32
    else:
        return xp.float64


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_hat_skew_symmetric(backend, batch_dims, precision, pass_xp):
    """Test that hat produces skew-symmetric matrices."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    dtype = get_dtype(backend, precision)

    # Create random tangent vectors
    w = 0.1 * xp.asarray(np.random.randn(*batch_dims, 3), dtype=dtype)
    W = SO3.hat(w, **xp_kwargs)

    assert W.shape == batch_dims + (3, 3)
    assert W.dtype == w.dtype

    W_np = np.array(W)
    # Properly transpose batch dimensions: swap last two axes
    W_T_np = np.transpose(W_np, axes=tuple(range(len(W.shape) - 2)) + (-1, -2))

    # Check that W = -W^T (skew-symmetric property)
    assert np.allclose(W_np, -W_T_np, atol=ATOL[precision])

    # Check that diagonal is zero
    diag_indices = np.arange(3)
    diagonal = W_np[..., diag_indices, diag_indices]
    assert np.allclose(diagonal, 0.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_hat_zero_vector(backend, batch_dims, precision, pass_xp):
    """Test hat operator on zero vector gives zero matrix."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    dtype = get_dtype(backend, precision)

    w = xp.zeros(batch_dims + (3,), dtype=dtype)
    W = SO3.hat(w, **xp_kwargs)

    assert W.shape == batch_dims + (3, 3)
    assert W.dtype == w.dtype

    W_np = np.array(W)
    assert np.allclose(W_np, 0.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_vee_hat_inverse(backend, batch_dims, precision, pass_xp):
    """Test that vee(hat(w)) = w."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    dtype = get_dtype(backend, precision)

    # Create random tangent vectors
    w = 0.1 * xp.asarray(np.random.randn(*batch_dims, 3), dtype=dtype)
    w_recovered = SO3.vee(SO3.hat(w, **xp_kwargs), **xp_kwargs)

    assert w_recovered.shape == batch_dims + (3,)
    assert w_recovered.dtype == w.dtype

    w_np = np.array(w)
    w_recovered_np = np.array(w_recovered)
    assert np.allclose(w_np, w_recovered_np, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_hat_vee_inverse_skew_symmetric(backend, batch_dims, precision, pass_xp):
    """Test that hat(vee(W)) = W for skew-symmetric W."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    dtype = get_dtype(backend, precision)

    # Create random tangent vector, then convert to skew-symmetric matrix
    w = 0.1 * xp.asarray(np.random.randn(*batch_dims, 3), dtype=dtype)
    W = SO3.hat(w, **xp_kwargs)
    W_recovered = SO3.hat(SO3.vee(W, **xp_kwargs), **xp_kwargs)

    assert W_recovered.shape == batch_dims + (3, 3)
    assert W_recovered.dtype == W.dtype

    W_np = np.array(W)
    W_recovered_np = np.array(W_recovered)
    assert np.allclose(W_np, W_recovered_np, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_vee_zero_matrix(backend, batch_dims, precision, pass_xp):
    """Test vee operator on zero matrix gives zero vector."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    dtype = get_dtype(backend, precision)

    W = xp.zeros(batch_dims + (3, 3), dtype=dtype)
    w = SO3.vee(W, **xp_kwargs)

    assert w.shape == batch_dims + (3,)
    assert w.dtype == W.dtype

    w_np = np.array(w)
    assert np.allclose(w_np, 0.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_hat_vee_numerical_precision(backend, batch_dims, precision, pass_xp):
    """Test numerical precision of hat/vee operations."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    dtype = get_dtype(backend, precision)

    if precision == 16:
        atol = 1e-3
    elif precision == 32:
        atol = 1e-6
    else:
        atol = 1e-14

    # Test with mathematically meaningful values
    pi = xp.asarray(np.pi, dtype=dtype)
    e = xp.asarray(np.e, dtype=dtype)
    sqrt2 = xp.asarray(np.sqrt(2), dtype=dtype)

    if len(batch_dims) == 0:
        w = xp.stack([pi, e, sqrt2])
    else:
        # For batch dimensions, create appropriate shape
        w_base = xp.stack([pi, e, sqrt2])
        w = xp.broadcast_to(w_base, batch_dims + (3,))

    w_recovered = SO3.vee(SO3.hat(w, **xp_kwargs), **xp_kwargs)

    assert w_recovered.dtype == w.dtype

    if precision >= 16:
        w_np = np.array(w)
        w_recovered_np = np.array(w_recovered)
        assert np.allclose(w_np, w_recovered_np, atol=atol)


def test_hat_specific_values():
    """Test hat operator with specific known values."""
    w = np.array([1.0, 2.0, 3.0])
    W = SO3.hat(w)

    expected = np.array([[0.0, -3.0, 2.0], [3.0, 0.0, -1.0], [-2.0, 1.0, 0.0]])

    W_np = np.array(W)
    assert np.allclose(W_np, expected)


def test_vee_specific_values():
    """Test vee operator with specific known values."""
    W = np.array([[0.0, -3.0, 2.0], [3.0, 0.0, -1.0], [-2.0, 1.0, 0.0]])
    w = SO3.vee(W)

    expected = np.array([1.0, 2.0, 3.0])

    w_np = np.array(w)
    assert np.allclose(w_np, expected)


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_hat_vee_large_values(backend, pass_xp):
    """Test hat and vee with large values."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)

    w = xp.asarray([1e6, -1e6, 1e6], dtype=xp.float32)
    W = SO3.hat(w, **xp_kwargs)

    assert W.dtype == w.dtype

    # Should still be skew-symmetric
    W_np = np.array(W)
    W_T_np = np.transpose(W_np, axes=(-1, -2))  # Just swap last two axes
    assert np.allclose(W_np, -W_T_np)


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_hat_vee_small_values(backend, pass_xp):
    """Test hat and vee with very small values."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)

    w = xp.asarray([1e-10, -1e-10, 1e-10], dtype=xp.float32)
    W = SO3.hat(w, **xp_kwargs)

    assert W.dtype == w.dtype

    # Should still be skew-symmetric
    W_np = np.array(W)
    W_T_np = np.transpose(W_np, axes=(-1, -2))  # Just swap last two axes
    assert np.allclose(W_np, -W_T_np, atol=1e-15)



import numpy as np
import pytest
from conftest import ATOL, TEST_BACKENDS, TEST_BATCH_DIMS, TEST_PASS_XP, TEST_PRECISIONS, get_xp_kwargs

from nanomanifold import SE3, SO3


def get_dtype(backend, precision):
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend)
    if precision == 16:
        return xp.float16
    elif precision == 32:
        return xp.float32
    else:
        return xp.float64


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_hat_correct_structure(backend, batch_dims, precision, pass_xp):
    """Test that hat produces correct 4x4 structure: top-left is SO3.hat, top-right is rho, bottom row is zeros."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend)
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    dtype = get_dtype(backend, precision)

    v = 0.1 * xp.asarray(np.random.randn(*batch_dims, 6), dtype=dtype)
    M = SE3.hat(v, **xp_kwargs)

    assert M.shape == batch_dims + (4, 4)
    assert M.dtype == v.dtype

    M_np = np.array(M)

    # Bottom row should be zeros
    assert np.allclose(M_np[..., 3, :], 0.0, atol=ATOL[precision])

    # Top-left 3x3 should match SO3.hat(omega)
    omega = v[..., :3]
    so3_hat_result = SO3.hat(omega, **xp_kwargs)
    so3_hat_np = np.array(so3_hat_result)
    assert np.allclose(M_np[..., :3, :3], so3_hat_np, atol=ATOL[precision])

    # Top-right column should be rho
    rho_np = np.array(v[..., 3:6])
    assert np.allclose(M_np[..., :3, 3], rho_np, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_hat_zero_input(backend, batch_dims, precision, pass_xp):
    """Test hat on zero vector gives zero matrix."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend)
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    dtype = get_dtype(backend, precision)

    v = xp.zeros(batch_dims + (6,), dtype=dtype)
    M = SE3.hat(v, **xp_kwargs)

    assert M.shape == batch_dims + (4, 4)
    M_np = np.array(M)
    assert np.allclose(M_np, 0.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_vee_hat_inverse(backend, batch_dims, precision, pass_xp):
    """Test that vee(hat(v)) == v."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend)
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    dtype = get_dtype(backend, precision)

    v = 0.1 * xp.asarray(np.random.randn(*batch_dims, 6), dtype=dtype)
    v_recovered = SE3.vee(SE3.hat(v, **xp_kwargs), **xp_kwargs)

    assert v_recovered.shape == batch_dims + (6,)
    assert v_recovered.dtype == v.dtype

    v_np = np.array(v)
    v_recovered_np = np.array(v_recovered)
    assert np.allclose(v_np, v_recovered_np, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_hat_vee_inverse(backend, batch_dims, precision, pass_xp):
    """Test that hat(vee(M)) == M for valid se(3) matrices."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend)
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    dtype = get_dtype(backend, precision)

    v = 0.1 * xp.asarray(np.random.randn(*batch_dims, 6), dtype=dtype)
    M = SE3.hat(v, **xp_kwargs)
    M_recovered = SE3.hat(SE3.vee(M, **xp_kwargs), **xp_kwargs)

    assert M_recovered.shape == batch_dims + (4, 4)
    assert M_recovered.dtype == M.dtype

    M_np = np.array(M)
    M_recovered_np = np.array(M_recovered)
    assert np.allclose(M_np, M_recovered_np, atol=ATOL[precision])


def test_hat_specific_values():
    """Test hat with specific known values."""
    v = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
    M = SE3.hat(v)
    M_np = np.array(M)

    expected = np.array(
        [
            [0.0, -3.0, 2.0, 4.0],
            [3.0, 0.0, -1.0, 5.0],
            [-2.0, 1.0, 0.0, 6.0],
            [0.0, 0.0, 0.0, 0.0],
        ]
    )
    assert np.allclose(M_np, expected)


def test_vee_specific_values():
    """Test vee with specific known values."""
    M = np.array(
        [
            [0.0, -3.0, 2.0, 4.0],
            [3.0, 0.0, -1.0, 5.0],
            [-2.0, 1.0, 0.0, 6.0],
            [0.0, 0.0, 0.0, 0.0],
        ]
    )
    v = SE3.vee(M)
    v_np = np.array(v)

    expected = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
    assert np.allclose(v_np, expected)


@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_hat_differentiability_torch(batch_dims):
    """Test that hat function is differentiable with PyTorch."""
    torch = pytest.importorskip("torch")
    dtype = torch.float64

    v = torch.randn(*batch_dims, 6, dtype=dtype, requires_grad=True)

    def f(v):
        return SE3.hat(v).sum()

    assert torch.autograd.gradcheck(f, (v,), eps=1e-6, atol=1e-5)


@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_vee_differentiability_torch(batch_dims):
    """Test that vee function is differentiable with PyTorch."""
    torch = pytest.importorskip("torch")
    dtype = torch.float64

    # Create a valid se(3) matrix
    v = torch.randn(*batch_dims, 6, dtype=dtype)
    M = SE3.hat(v).detach().requires_grad_(True)

    def f(M):
        return SE3.vee(M).sum()

    assert torch.autograd.gradcheck(f, (M,), eps=1e-6, atol=1e-5)


@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_hat_vee_jittability_jax(batch_dims):
    """Test that hat and vee are JIT-compatible with JAX."""
    jax = pytest.importorskip("jax")
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name("jax")

    v = xp.asarray(np.random.randn(*batch_dims, 6).astype("float32"))

    @jax.jit
    def jit_hat(v):
        return SE3.hat(v)

    @jax.jit
    def jit_vee(M):
        return SE3.vee(M)

    M_jit = jit_hat(v)
    v_jit = jit_vee(M_jit)

    M_non_jit = SE3.hat(v)
    v_non_jit = SE3.vee(M_non_jit)

    assert jax.numpy.allclose(M_jit, M_non_jit, atol=1e-6)
    assert jax.numpy.allclose(v_jit, v_non_jit, atol=1e-6)



import numpy as np
import pytest
from conftest import ATOL, TEST_BACKENDS, TEST_BATCH_DIMS, TEST_PASS_XP, TEST_PRECISIONS, get_xp_kwargs, identity_se3, random_se3

from nanomanifold import SE3


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_log_identity(backend, batch_dims, precision, pass_xp):
    """Test that log of identity SE3 transformation gives zero tangent vector."""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    identity = identity_se3(batch_dims=batch_dims, backend=backend, precision=precision)
    log_result = SE3.log(identity, **xp_kwargs)

    assert log_result.shape == batch_dims + (6,)

    log_np = np.array(log_result)
    assert np.allclose(log_np, 0.0, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_exp_zero(backend, batch_dims, precision, pass_xp):
    """Test that exp of zero tangent vector gives identity SE3 transformation."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)

    if precision == 16:
        dtype = xp.float16
    elif precision == 32:
        dtype = xp.float32
    else:
        dtype = xp.float64

    zero_tangent = xp.zeros(batch_dims + (6,), dtype=dtype)
    exp_result = SE3.exp(zero_tangent, **xp_kwargs)

    identity = identity_se3(batch_dims=batch_dims, backend=backend, precision=precision)

    assert exp_result.shape == identity.shape

    exp_np = np.array(exp_result)
    identity_np = np.array(identity)

    # Check quaternion equivalence (accounting for q/-q ambiguity)
    q_exp = exp_np[..., :4]
    q_identity = identity_np[..., :4]
    dot_products = np.sum(q_exp * q_identity, axis=-1)
    assert np.allclose(np.abs(dot_products), 1.0, atol=ATOL[precision])

    t_exp = exp_np[..., 4:7]
    t_identity = identity_np[..., 4:7]
    assert np.allclose(t_exp, t_identity, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_log_exp_inverse(backend, batch_dims, precision, pass_xp):
    """Test that exp(log(se3)) = se3 for random SE3 transformations."""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    se3 = random_se3(batch_dims=batch_dims, backend=backend, precision=precision)

    log_se3 = SE3.log(se3, **xp_kwargs)
    exp_log_se3 = SE3.exp(log_se3, **xp_kwargs)

    assert exp_log_se3.shape == se3.shape

    se3_np = np.array(se3)
    result_np = np.array(exp_log_se3)

    # Check quaternion equivalence (q and -q represent same rotation)
    q_orig = se3_np[..., :4]
    q_result = result_np[..., :4]
    dot_products = np.sum(q_orig * q_result, axis=-1)
    assert np.allclose(np.abs(dot_products), 1.0, atol=ATOL[precision])

    t_orig = se3_np[..., 4:7]
    t_result = result_np[..., 4:7]
    assert np.allclose(t_orig, t_result, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_exp_log_inverse(backend, batch_dims, precision, pass_xp):
    """Test that log(exp(tangent)) = tangent for small tangent vectors."""
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name(backend.replace("jax", "jax"))
    xp_kwargs = get_xp_kwargs(backend, pass_xp)

    shape = batch_dims + (6,)
    tangent_vec_np = 0.5 * np.random.normal(0, 1, size=shape).astype(f"float{precision}")
    tangent_vec = xp.asarray(tangent_vec_np)

    exp_vec = SE3.exp(tangent_vec, **xp_kwargs)
    log_exp_vec = SE3.log(exp_vec, **xp_kwargs)

    assert log_exp_vec.shape == tangent_vec.shape

    original_np = np.array(tangent_vec)
    result_np = np.array(log_exp_vec)
    assert np.allclose(original_np, result_np, atol=ATOL[precision])


def test_log_specific_transformations():
    """Test log for known SE3 transformations."""
    se3_translation = np.array([1.0, 0.0, 0.0, 0.0, 1.0, 2.0, 3.0])

    sqrt2_inv = 1.0 / np.sqrt(2.0)
    se3_rotation = np.array([sqrt2_inv, 0.0, 0.0, sqrt2_inv, 0.0, 0.0, 0.0])

    se3_combined = np.array([sqrt2_inv, 0.0, 0.0, sqrt2_inv, 1.0, 2.0, 3.0])

    log_translation = SE3.log(se3_translation)
    log_translation_np = np.array(log_translation)
    expected_translation = [0.0, 0.0, 0.0, 1.0, 2.0, 3.0]  # [ω=0, ρ=t]
    assert np.allclose(log_translation_np, expected_translation, atol=1e-6)

    log_rotation = SE3.log(se3_rotation)
    log_rotation_np = np.array(log_rotation)
    expected_rotation = [0.0, 0.0, np.pi / 2, 0.0, 0.0, 0.0]  # [ω=z*π/2, ρ=0]
    assert np.allclose(log_rotation_np, expected_rotation, atol=1e-6)

    log_combined = SE3.log(se3_combined)
    log_combined_np = np.array(log_combined)
    # For combined, the translation part is transformed by the V^(-1) matrix
    # We mainly test that it's reasonable and invertible
    assert log_combined_np.shape == (6,)
    assert np.allclose(log_combined_np[:3], [0.0, 0.0, np.pi / 2], atol=1e-6)  # Rotation part unchanged


def test_exp_specific_tangents():
    """Test exp for known tangent vectors."""
    zero_tangent = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])

    omega_z90 = np.array([0.0, 0.0, np.pi / 2, 0.0, 0.0, 0.0])

    rho_translation = np.array([0.0, 0.0, 0.0, 1.0, 2.0, 3.0])

    exp_zero = SE3.exp(zero_tangent)
    exp_zero_np = np.array(exp_zero)
    expected_identity = [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    # Check quaternion equivalence and exact translation
    dot_product = np.sum(exp_zero_np[:4] * expected_identity[:4])
    assert np.allclose(np.abs(dot_product), 1.0, atol=1e-6)
    assert np.allclose(exp_zero_np[4:], expected_identity[4:], atol=1e-6)

    exp_rotation = SE3.exp(omega_z90)
    exp_rotation_np = np.array(exp_rotation)
    sqrt2_inv = 1.0 / np.sqrt(2.0)
    expected_rotation = [sqrt2_inv, 0.0, 0.0, sqrt2_inv, 0.0, 0.0, 0.0]
    dot_product = np.sum(exp_rotation_np[:4] * expected_rotation[:4])
    assert np.allclose(np.abs(dot_product), 1.0, atol=1e-6)
    assert np.allclose(exp_rotation_np[4:], expected_rotation[4:], atol=1e-6)

    exp_translation = SE3.exp(rho_translation)
    exp_translation_np = np.array(exp_translation)
    expected_translation = [1.0, 0.0, 0.0, 0.0, 1.0, 2.0, 3.0]
    dot_product = np.sum(exp_translation_np[:4] * expected_translation[:4])
    assert np.allclose(np.abs(dot_product), 1.0, atol=1e-6)
    assert np.allclose(exp_translation_np[4:], expected_translation[4:], atol=1e-6)


def test_log_exp_matrix_consistency():
    """Test that log/exp is consistent with matrix conversion."""
    se3 = random_se3(batch_dims=(), backend="numpy", precision=32)

    matrix = SE3.to_matrix(se3)
    se3_from_matrix = SE3.from_matrix(matrix)

    log_orig = SE3.log(se3)
    log_from_matrix = SE3.log(se3_from_matrix)

    log_orig_np = np.array(log_orig)
    log_from_matrix_np = np.array(log_from_matrix)

    assert np.allclose(log_orig_np, log_from_matrix_np, atol=1e-6)


def test_log_exp_composition():
    """Test that log and exp compose properly with other operations."""
    se3_1 = random_se3(batch_dims=(), backend="numpy", precision=32)
    se3_2 = random_se3(batch_dims=(), backend="numpy", precision=32)

    se3_product = SE3.multiply(se3_1, se3_2)

    log_product = SE3.log(se3_product)

    exp_log_product = SE3.exp(log_product)

    product_np = np.array(se3_product)
    result_np = np.array(exp_log_product)

    # Check quaternion equivalence
    q_prod = product_np[:4]
    q_result = result_np[:4]
    dot_product = np.sum(q_prod * q_result)
    assert np.allclose(np.abs(dot_product), 1.0, atol=1e-6)

    # Check translation equivalence
    t_prod = product_np[4:]
    t_result = result_np[4:]
    assert np.allclose(t_prod, t_result, atol=1e-6)


@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_log_differentiability_torch(batch_dims):
    """Test that log function is differentiable with PyTorch."""
    torch = pytest.importorskip("torch")
    # Use double precision for gradient checking
    dtype = torch.float64

    # Random SE3 input (avoiding singularities)
    se3 = random_se3(batch_dims=batch_dims, backend="torch").to(dtype).requires_grad_(True)

    def f_log(se3):
        return SE3.log(se3).sum()

    assert torch.autograd.gradcheck(f_log, (se3,), eps=1e-6, atol=1e-5)


@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_exp_differentiability_torch(batch_dims):
    """Test that exp function is differentiable with PyTorch."""
    torch = pytest.importorskip("torch")
    dtype = torch.float64

    shape = batch_dims + (6,)
    tangent_vec_np = 0.5 * np.random.normal(0, 1, size=shape).astype("float64")
    tangent_vec = torch.tensor(tangent_vec_np, dtype=dtype, requires_grad=True)

    def f_exp(v):
        return SE3.exp(v).sum()

    assert torch.autograd.gradcheck(f_exp, (tangent_vec,), eps=1e-6, atol=1e-5)


@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_log_exp_jittability_jax(batch_dims):
    """Test that log and exp functions are JIT-compatible with JAX."""
    jax = pytest.importorskip("jax")
    common = pytest.importorskip("nanomanifold.common")
    xp = common.get_namespace_by_name("jax")

    se3 = random_se3(batch_dims=batch_dims, backend="jax", precision=32)
    shape = batch_dims + (6,)
    tangent_np = 0.1 * np.random.normal(0, 1, size=shape).astype("float32")
    tangent = xp.asarray(tangent_np)

    @jax.jit
    def jit_log(se3):
        return SE3.log(se3)

    @jax.jit
    def jit_exp(v):
        return SE3.exp(v)

    log_result_jit = jit_log(se3)
    exp_result_jit = jit_exp(tangent)

    log_result_non_jit = SE3.log(se3)
    exp_result_non_jit = SE3.exp(tangent)

    assert jax.numpy.allclose(log_result_jit, log_result_non_jit, atol=1e-6)
    assert jax.numpy.allclose(exp_result_jit, exp_result_non_jit, atol=1e-6)

    assert log_result_jit.shape == batch_dims + (6,)
    assert exp_result_jit.shape == batch_dims + (7,)

    q_result = exp_result_jit[..., :4]
    q_norm = jax.numpy.linalg.norm(q_result, axis=-1)
    assert jax.numpy.allclose(q_norm, 1.0, atol=1e-6)


def test_log_exp_broadcasting():
    """Test broadcasting between different batch sizes."""
    se3_single = random_se3(batch_dims=(), backend="numpy", precision=32)

    tangent_batch = 0.1 * np.random.normal(0, 1, size=(3, 6)).astype("float32")

    log_single = SE3.log(se3_single)
    assert log_single.shape == (6,)

    exp_batch = SE3.exp(tangent_batch)
    assert exp_batch.shape == (3, 7)

    exp_batch_np = np.array(exp_batch)
    assert np.all(np.isfinite(exp_batch_np))

    q_norms = np.linalg.norm(exp_batch_np[:, :4], axis=-1)
    assert np.allclose(q_norms, 1.0, atol=1e-6)


def test_log_exp_edge_cases():
    """Test edge cases and numerical stability."""
    small_tangent = np.array([1e-8, 1e-8, 1e-8, 1e-8, 1e-8, 1e-8])
    se3_small = SE3.exp(small_tangent)
    log_small = SE3.log(se3_small)

    small_tangent_np = np.array(small_tangent)
    log_small_np = np.array(log_small)
    assert np.allclose(small_tangent_np, log_small_np, atol=1e-12)

    large_tangent = np.array([np.pi - 1e-6, 0, 0, 1.0, 2.0, 3.0])
    se3_large = SE3.exp(large_tangent)
    log_large = SE3.log(se3_large)

    large_tangent_np = np.array(large_tangent)
    log_large_np = np.array(log_large)
    assert np.allclose(large_tangent_np, log_large_np, atol=1e-5)

    pure_rotation_tangent = np.array([0.5, 0.3, 0.1, 0.0, 0.0, 0.0])
    se3_pure_rot = SE3.exp(pure_rotation_tangent)
    log_pure_rot = SE3.log(se3_pure_rot)

    pure_rotation_np = np.array(pure_rotation_tangent)
    log_pure_rot_np = np.array(log_pure_rot)
    assert np.allclose(pure_rotation_np, log_pure_rot_np, atol=1e-6)

    pure_translation_tangent = np.array([0.0, 0.0, 0.0, 1.0, 2.0, 3.0])
    se3_pure_trans = SE3.exp(pure_translation_tangent)
    log_pure_trans = SE3.log(se3_pure_trans)

    pure_translation_np = np.array(pure_translation_tangent)
    log_pure_trans_np = np.array(log_pure_trans)
    assert np.allclose(pure_translation_np, log_pure_trans_np, atol=1e-6)



"""6D rotation representation conversions for SO(3)."""

from types import ModuleType
from typing import Any

from jaxtyping import Float

from nanomanifold import common
from nanomanifold.common import get_namespace

from .matrix import from_matrix, to_matrix


def _normalize(vec: Float[Any, "... 3"], xp) -> Float[Any, "... 3"]:
    eps = common.safe_eps(vec.dtype, xp)
    norm = xp.linalg.norm(vec, axis=-1, keepdims=True)
    safe_norm = xp.where(norm < eps, xp.ones_like(norm), norm)
    return vec / safe_norm


def _cross(a: Float[Any, "... 3"], b: Float[Any, "... 3"], xp) -> Float[Any, "... 3"]:
    ax, ay, az = a[..., 0], a[..., 1], a[..., 2]
    bx, by, bz = b[..., 0], b[..., 1], b[..., 2]
    return xp.stack([ay * bz - az * by, az * bx - ax * bz, ax * by - ay * bx], axis=-1)


def _from_6d_to_matrix(d6: Float[Any, "... 6"], xp) -> Float[Any, "... 3 3"]:
    a1 = d6[..., 0:3]
    a2 = d6[..., 3:6]

    b1 = _normalize(a1, xp)
    dot = xp.sum(b1 * a2, axis=-1, keepdims=True)
    b2 = _normalize(a2 - dot * b1, xp)
    b3 = _cross(b1, b2, xp)

    return xp.stack([b1, b2, b3], axis=-1)


def to_6d(q: Float[Any, "... 4"], *, xp: ModuleType | None = None) -> Float[Any, "... 6"]:
    """Convert quaternion rotation to 6D representation.

    The 6D representation is formed by concatenating the first two columns of
    the rotation matrix.
    """
    if xp is None:
        xp = get_namespace(q)
    R = to_matrix(q, xp=xp)
    return xp.concatenate([R[..., :, 0], R[..., :, 1]], axis=-1)


def from_6d(d6: Float[Any, "... 6"], *, xp: ModuleType | None = None) -> Float[Any, "... 4"]:
    """Convert 6D rotation representation to a quaternion.

    Applies Gram-Schmidt orthonormalization to ensure a valid rotation matrix.
    """
    if xp is None:
        xp = get_namespace(d6)
    R = _from_6d_to_matrix(d6, xp)
    return from_matrix(R, xp=xp)



from . import conversions
from .distance import distance
from .exp import exp
from .hat import hat
from .inverse import inverse
from .log import log
from .multiply import multiply
from .primitives.axis_angle import from_axis_angle, to_axis_angle
from .primitives.euler import from_euler, to_euler
from .primitives.matrix import from_matrix, to_matrix
from .primitives.quaternion import canonicalize, from_quat_xyzw, to_quat_xyzw
from .primitives.sixd import from_6d, to_6d
from .random import random
from .rotate_points import rotate_points
from .slerp import slerp
from .vee import vee
from .weighted_mean import mean, weighted_mean

__all__ = [
    "conversions",
    "to_axis_angle",
    "from_axis_angle",
    "to_euler",
    "from_euler",
    "to_matrix",
    "from_matrix",
    "from_quat_xyzw",
    "to_quat_xyzw",
    "to_6d",
    "from_6d",
    "canonicalize",
    "rotate_points",
    "inverse",
    "multiply",
    "distance",
    "log",
    "exp",
    "hat",
    "vee",
    "slerp",
    "weighted_mean",
    "mean",
    "random",
]



import numpy as np
import pytest
from conftest import (
    ATOL,
    TEST_BACKENDS,
    TEST_BATCH_DIMS,
    TEST_PASS_XP,
    TEST_PRECISIONS,
    get_xp_kwargs,
    identity_se3,
    random_points,
    random_se3,
)

from nanomanifold import SE3, SO3


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_transform_points_identity(backend, batch_dims, precision, pass_xp):
    """Test that identity SE3 transformation doesn't change points."""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    # Create identity SE3 transformation [1, 0, 0, 0, 0, 0, 0]
    identity = identity_se3(batch_dims=batch_dims, backend=backend, precision=precision)

    # Generate random points
    points = random_points(batch_dims=batch_dims, n_points=10, backend=backend, precision=precision)

    # Apply transformation
    transformed_points = SE3.transform_points(identity, points, **xp_kwargs)

    assert transformed_points.dtype == points.dtype
    assert transformed_points.shape == points.shape

    # Convert to numpy arrays and compare
    points_np = np.array(points)
    transformed_points_np = np.array(transformed_points)

    assert np.allclose(points_np, transformed_points_np, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_transform_points_inverse(backend, batch_dims, precision, pass_xp):
    """Test that transforming by SE3 then by SE3^(-1) gives identity."""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    # Create random SE3 transformation
    se3 = random_se3(batch_dims=batch_dims, backend=backend, precision=precision)

    # Create inverse SE3 transformation using SE3.inverse()
    se3_inv = SE3.inverse(se3, **xp_kwargs)

    # Generate random points
    points = random_points(batch_dims=batch_dims, n_points=10, backend=backend, precision=precision)

    # Transform by se3 then by se3^(-1)
    transformed_points = SE3.transform_points(se3, points, **xp_kwargs)
    restored_points = SE3.transform_points(se3_inv, transformed_points, **xp_kwargs)

    assert restored_points.dtype == points.dtype
    assert restored_points.shape == points.shape

    # Convert to numpy arrays and compare
    points_np = np.array(points)
    restored_points_np = np.array(restored_points)

    # Use more lenient tolerance for complex batch dimensions due to accumulated floating-point error
    assert np.allclose(points_np, restored_points_np, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("precision", TEST_PRECISIONS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_transform_points_matrix_equivalence(backend, batch_dims, precision, pass_xp):
    """Test that SE3 point transformation matches matrix transformation."""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    # Create random SE3 transformation
    se3 = random_se3(batch_dims=batch_dims, backend=backend, precision=precision)

    # Generate random points
    points = random_points(batch_dims=batch_dims, n_points=10, backend=backend, precision=precision)

    # Transform points using SE3.transform_points
    transformed_points = SE3.transform_points(se3, points, **xp_kwargs)

    # Transform points using matrix approach
    matrix = SE3.to_matrix(se3, **xp_kwargs)
    matrix_np = np.array(matrix)
    points_np = np.array(points)

    # Add homogeneous coordinate (w=1)
    ones = np.ones(points_np.shape[:-1] + (1,))
    points_homogeneous = np.concatenate([points_np, ones], axis=-1)  # (..., N, 4)

    # Apply matrix transformation
    if matrix_np.ndim == 2:  # Single transformation
        # For single transformation: (N,4) @ (4,4)^T -> (N,4)
        transformed_homogeneous = np.matmul(points_homogeneous, matrix_np.T)
    else:  # Batched transformations
        # Einstein summation for batched matrix multiplication
        transformed_homogeneous = np.einsum("...ij,...nj->...ni", matrix_np, points_homogeneous)

    # Remove homogeneous coordinate
    transformed_matrix = transformed_homogeneous[..., :3]

    assert transformed_points.dtype == points.dtype
    assert transformed_points.shape == points.shape

    # Convert to numpy arrays and compare
    transformed_points_np = np.array(transformed_points)

    assert np.allclose(transformed_points_np, transformed_matrix, atol=ATOL[precision])


@pytest.mark.parametrize("backend", TEST_BACKENDS)
@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
@pytest.mark.parametrize("pass_xp", TEST_PASS_XP)
def test_transform_points_scipy(backend, batch_dims, pass_xp):
    """Test point transformation against scipy spatial transform."""
    xp_kwargs = get_xp_kwargs(backend, pass_xp)
    pytest.importorskip("scipy.spatial")
    from scipy.spatial.transform import Rotation as R

    # Create random SE3 transformation
    se3 = random_se3(batch_dims=batch_dims, backend=backend, precision=32)

    # Generate random points
    points = random_points(batch_dims=batch_dims, n_points=10, backend=backend, precision=32)

    # Transform using nanomanifold
    transformed_points = SE3.transform_points(se3, points, **xp_kwargs)

    # Convert to numpy for scipy
    se3_np = np.array(se3)
    points_np = np.array(points)

    # Extract quaternion and translation
    q_np = se3_np[..., :4]
    t_np = se3_np[..., 4:7]

    # Convert from [w, x, y, z] to scipy's [x, y, z, w] format
    q_scipy = np.concatenate([q_np[..., 1:4], q_np[..., 0:1]], axis=-1)

    # Reshape for scipy processing
    q_flat = q_scipy.reshape(-1, 4)
    t_flat = t_np.reshape(-1, 3)
    points_flat = points_np.reshape(-1, points_np.shape[-2], 3)

    # Transform using scipy
    transformed_scipy = []
    for i in range(len(q_flat)):
        r = R.from_quat(q_flat[i : i + 1])
        # Apply rotation then translation: p' = R*p + t
        rotated = r.apply(points_flat[i])
        transformed = rotated + t_flat[i : i + 1, None, :]
        transformed_scipy.append(transformed)

    transformed_scipy = np.array(transformed_scipy).reshape(points_np.shape)

    # Convert to numpy for comparison
    transformed_points_np = np.array(transformed_points)

    assert np.allclose(transformed_points_np, transformed_scipy, atol=1e-6)


@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_transform_points_differentiability_torch(batch_dims):
    """Test that transform_points function is differentiable with PyTorch."""
    torch = pytest.importorskip("torch")
    # Use double precision for gradient checking as recommended by PyTorch
    dtype = torch.float64

    # Random SE3 input
    se3 = random_se3(batch_dims=batch_dims, backend="torch").to(dtype).requires_grad_(True)
    points = random_points(batch_dims=batch_dims, n_points=5, backend="torch").to(dtype)

    # Check gradients of SE3.transform_points with respect to SE3 transformation
    def f(se3):
        return SE3.transform_points(se3, points)

    assert torch.autograd.gradcheck(f, (se3,), eps=1e-6, atol=1e-5)


@pytest.mark.parametrize("batch_dims", TEST_BATCH_DIMS)
def test_transform_points_jittability_jax(batch_dims):
    """Test that transform_points function is JIT-compatible with JAX."""
    jax = pytest.importorskip("jax")

    # Create test SE3 transformation and points
    se3 = random_se3(batch_dims=batch_dims, backend="jax", precision=32)
    points = random_points(batch_dims=batch_dims, n_points=5, backend="jax", precision=32)

    # Define JIT-compiled function
    @jax.jit
    def jit_transform_points(se3, points):
        return SE3.transform_points(se3, points)

    # Test that JIT compilation works and compare with non-JIT
    result_jit = jit_transform_points(se3, points)
    result_non_jit = SE3.transform_points(se3, points)

    # Verify results match between JIT and non-JIT
    assert jax.numpy.allclose(result_jit, result_non_jit, atol=1e-6)

    # Verify result has correct shape
    expected_shape = batch_dims + (5, 3)
    assert result_jit.shape == expected_shape


def test_transform_points_specific_transformations():
    """Test transformation of specific known cases."""
    # Pure translation: [1, 0, 0, 0, 1, 2, 3]
    se3_translation = np.array([1.0, 0.0, 0.0, 0.0, 1.0, 2.0, 3.0])
    points = np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]])  # Origin and unit x

    transformed = SE3.transform_points(se3_translation, points)
    transformed_np = np.array(transformed)

    # Expected: points translated by (1, 2, 3)
    expected = np.array([[1.0, 2.0, 3.0], [2.0, 2.0, 3.0]])
    assert np.allclose(transformed_np, expected, atol=1e-6)

    # 90-degree rotation around z-axis with translation
    sqrt2_inv = 1.0 / np.sqrt(2.0)
    se3_rot_trans = np.array([sqrt2_inv, 0.0, 0.0, sqrt2_inv, 1.0, 0.0, 0.0])
    points = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])  # Unit x and unit y

    transformed = SE3.transform_points(se3_rot_trans, points)
    transformed_np = np.array(transformed)

    # Expected: 90° rotation around z + translation (1,0,0)
    # (1,0,0) -> (0,1,0) + (1,0,0) = (1,1,0)
    # (0,1,0) -> (-1,0,0) + (1,0,0) = (0,0,0)
    expected = np.array([[1.0, 1.0, 0.0], [0.0, 0.0, 0.0]])
    assert np.allclose(transformed_np, expected, atol=1e-6)


def test_transform_points_composition():
    """Test that composing transformations gives same result as composed transformation."""
    # Create two random SE3 transformations
    se3_1 = random_se3(batch_dims=(), backend="numpy", precision=32)
    se3_2 = random_se3(batch_dims=(), backend="numpy", precision=32)

    # Compose them
    se3_composed = SE3.multiply(se3_1, se3_2)

    # Create random points
    points = random_points(batch_dims=(), n_points=5, backend="numpy", precision=32)

    # Method 1: Transform with composed transformation
    result_composed = SE3.transform_points(se3_composed, points)

    # Method 2: Transform sequentially (se3_2 first, then se3_1)
    intermediate = SE3.transform_points(se3_2, points)
    result_sequential = SE3.transform_points(se3_1, intermediate)

    # Results should match
    result_composed_np = np.array(result_composed)
    result_sequential_np = np.array(result_sequential)

    assert np.allclose(result_composed_np, result_sequential_np, atol=1e-6)


def test_transform_points_pure_rotation():
    """Test that SE3 with zero translation matches SO3 rotation."""

    # Create random quaternion
    quat = np.array([0.707, 0.707, 0.0, 0.0])  # 90° around x-axis
    quat = quat / np.linalg.norm(quat)  # Normalize

    # Create SE3 with zero translation
    se3_pure_rotation = np.concatenate([quat, np.zeros(3)])

    # Create test points
    points = np.array([[0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])  # Unit y and unit z

    # Transform using SE3
    se3_result = SE3.transform_points(se3_pure_rotation, points)

    # Transform using SO3
    so3_result = SO3.rotate_points(quat, points)

    # Results should match
    se3_result_np = np.array(se3_result)
    so3_result_np = np.array(so3_result)

    assert np.allclose(se3_result_np, so3_result_np, atol=1e-6)


def test_transform_points_broadcasting():
    """Test broadcasting between different SE3 and points batch sizes."""
    # Single SE3 transformation and batch of points
    se3_single = random_se3(batch_dims=(), backend="numpy", precision=32)
    points_batch = random_points(batch_dims=(3,), n_points=5, backend="numpy", precision=32)

    # Test broadcasting: single SE3 transforms multiple point sets
    result = SE3.transform_points(se3_single, points_batch)
    assert result.shape == (3, 5, 3)

    # Batch of SE3 transformations and single point set
    se3_batch = random_se3(batch_dims=(3,), backend="numpy", precision=32)
    points_single = random_points(batch_dims=(), n_points=5, backend="numpy", precision=32)

    # Test broadcasting: multiple SE3 transforms single point set
    result = SE3.transform_points(se3_batch, points_single)
    assert result.shape == (3, 5, 3)

    # Verify results are valid (finite numbers)
    result_np = np.array(result)
    assert np.all(np.isfinite(result_np))


def test_transform_points_single_point():
    """Test transformation of single point (edge case)."""
    # Create SE3 transformation
    se3 = np.array([0.707, 0.0, 0.707, 0.0, 1.0, 2.0, 3.0])  # 90° around y + translation

    # Single point
    point = np.array([[1.0, 0.0, 0.0]])  # Shape (1, 3)

    # Transform
    transformed = SE3.transform_points(se3, point)
    transformed_np = np.array(transformed)

    # Expected: 90° rotation around y transforms (1,0,0) -> (0,0,-1), then add (1,2,3)
    expected = np.array([[1.0, 2.0, 2.0]])
    assert transformed.shape == (1, 3)
    assert np.allclose(transformed_np, expected, atol=1e-6)
